## Automatically adapted for numpy.oldnumeric Sep 06, 2007 by alter_code1.py

from __future__ import print_function

 
# ImageD11_v0.4 Software for beamline ID11
# Copyright (C) 2005  Jon Wright
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


"""
Defines a class for correcting peak positions for spatial distortion
via a fit2d spline file
 
To think about doing - valid regions? What if someone uses a 1K spline
file for a 2K image etc?
"""
import logging, numpy, math
from scipy.interpolate import bisplev

def readfit2dfloats(filep, nfl):
    """
    Interprets a 5E14.7 formatted fortran line
    filep = file object (has readline method)
    nfl   = the number of floats to read
    """
    ret = []
    j = 0
    while j < nfl:
        i = 0
        thisline = filep.readline()
        # logging.debug("readfit2dfloats:"+thisline)
        while i < 5 * 14:
            # logging.debug(str(i)+ thisline[i:i+14])
            ret.append(float(thisline[i:i+14]) )
            j = j + 1
            i = i + 14
            if j == nfl: 
                break
    return ret

class correctorclass: #IGNORE:R0902
    """
    Applies a spatial distortion to a peak position using a fit2d splinefile
    """
    def __init__(self, argsplinefile, orientation="edf"):
        """
        Argument is the name of a fit2d spline file
        """
        self.splinefile = argsplinefile
        self.tolerance = 1e-5
        self.orientation = orientation
        self.pos_lut = None
        self.pixel_lut = None
        self.xmin = self.ymin = self.xmax = self.ymax = 0.0
        self.xsize = self.ysize = 1.0
        self.gridspacing = 0.0
        self.tck1 = None
        self.tck2 = None
        if self.splinefile is not None:
            self.readfit2dspline(self.splinefile)
        


    def correct(self, xin, yin):
        """
        Transform x,y in raw image coordinates into x,y of an
        idealised image. Returns a tuple (x,y), expects a
        pair of floats as arguments
        """
        if self.orientation == "edf":
            xcor = xin + bisplev(yin, xin, self.tck2)
            ycor = yin + bisplev(yin, xin, self.tck1)
        else: 
            # fit2d does a flip 
            raise Exception("Spline orientations must be edf, convert "
                            "your image to edf and remake the spline")
            # Unreachable code - we no longer accept this complexity
            # it means the spline file for ImageD11 bruker images
            # is not the same as for fit2d. 
            # xpos = self.xmax - xin
            # xcor = xin - bisplev(yin, xpos, self.tck2)
            # ycor = yin + bisplev(yin, xpos, self.tck1)
        return xcor, ycor

    def make_pixel_lut(self, dims):
        """
        Generate an x and y image which maps the array indices into
        floating point array indices (to be corrected for pixel size later)

        returns 
        FIXME - check they are the right way around
                add some sort of known splinefile testcase
        """
        # Cache the value in case of multiple calls
        if self.pixel_lut is None:
            x_im = numpy.outer(numpy.arange(dims[0]), numpy.ones(dims[1]))
            y_im = numpy.outer(numpy.ones(dims[0]), numpy.arange(dims[1]))
            # xcor is tck2
            x_im = numpy.add( x_im,
                              bisplev( numpy.arange(dims[1]),
                                         numpy.arange(dims[0]),
                                               self.tck2 ).T,
                              x_im)
            # ycor is tck1
            y_im = numpy.add( y_im,
                              bisplev( numpy.arange(dims[1]),
                                               numpy.arange(dims[0]),
                                               self.tck1 ).T,
                              y_im)
            self.pixel_lut = x_im, y_im
        return self.pixel_lut

    def make_pos_lut(self, dims):
        """
        Generate a look up table of pixel positions in microns
        # Cache the value in case of multiple calls
        returns ...
        """
        if self.pos_lut is None:
            if self.pixel_lut is None:
                self.make_pixel_lut(dims)                
            self.pos_lut = ( self.pixel_lut[0] * self.xsize, 
                             self.pixel_lut[1] * self.ysize )
        return self.pos_lut

    def distort(self, xin, yin):
        """
        Distort a pair of points xnew, ynew to find where they
        would be in a raw image

        Iterative algorithm...
        """
        yold = yin - bisplev(yin, xin, self.tck1)
        xold = xin - bisplev(yin, xin, self.tck2)
        # First guess, assumes distortion is constant
        ytmp = yin - bisplev(yold, xold, self.tck1)
        xtmp = xin - bisplev(yold, xold, self.tck2)
        # Second guess should be better
        error = math.sqrt((xtmp - xold) * (xtmp - xold) + 
                          (ytmp - yold) * (ytmp - yold)   )
        ntries = 0
        while error > self.tolerance:
            ntries = ntries + 1
            xold = xtmp
            yold = ytmp
            ytmp = yin - bisplev(yold, xold, self.tck1)
            xtmp = xin - bisplev(yold, xold, self.tck2)
            error = math.sqrt((xtmp - xold) * (xtmp - xold) + 
                              (ytmp - yold) * (ytmp - yold)   )
            # print error,xold,x,yold,y
            if ntries == 10:
                raise Exception("Error getting the inverse spline to converge")
        return xtmp, ytmp

    def test(self, xin, yin):
        """
        Checks that the correct and distort functions are indeed
        inversely related to each other
        """
        xtes, ytes = self.correct(xin, yin)
        xold, yold = self.distort(xtes, ytes)
        error = math.sqrt( (xin - xold) * (xin - xold) + 
                           (yin - yold) * (yin - yold))
        if error > self.tolerance:
            logging.error("Blobcorrector Test Failed!")
            raise Exception("Problem in correctorclass")






            # read the fit2d array into a tck tuple
    def readfit2dspline(self, name):
        """
        Reads a fit2d spline file into a scipy/fitpack tuple, tck
        A fairly long and dull routine...
        """
        fin = open(name, "r")
        # SPATIAL DISTORTION SPLINE INTERPOLATION COEFFICIENTS
        myline = fin.readline() 
        if myline[:7] != "SPATIAL":
            raise SyntaxError(name + \
                ": file does not seem to be a fit2d spline file")
        fin.readline() # BLANK LINE
        fin.readline() # VALID REGION
        myline = fin.readline() # the actual valid region, 
                               # assuming xmin,ymin,xmax,ymax
        logging.debug("xmin,ymin,xmax,ymax, read: "+myline)
        self.xmin, self.ymin, self.xmax, self.ymax = \
         [float(z) for z in myline.split()]
        myline = fin.readline() # BLANK
        myline = fin.readline() # GRID SPACING, X-PIXEL SIZE, Y-PIXEL SIZE
        myline = fin.readline()
        logging.debug("gridspace, xsize, ysize: "+myline)
        self.gridspacing, self.xsize, self.ysize = \
         [float(z) for z in  myline.split()]
        fin.readline() # BLANK
        fin.readline() # X-DISTORTION
        myline = fin.readline() # two integers nx1,ny1
        logging.debug("nx1, ny1 read: "+myline)
        nx1, ny1 = [int(z) for z in myline.split()]
        # Now follow fit2d formatted line 5E14.7
        tx1 = numpy.array(readfit2dfloats(fin, nx1), numpy.float32)
        ty1 = numpy.array(readfit2dfloats(fin, ny1), numpy.float32)
        cf1 = numpy.array(readfit2dfloats(fin, (nx1 - 4) * (ny1 - 4)),  
                          numpy.float32)
        fin.readline() #BLANK
        fin.readline() # Y-DISTORTION
        myline = fin.readline() # two integers nx2, ny2
        nx2 , ny2 = [int(z) for z in myline.split()]
        tx2 = numpy.array(readfit2dfloats(fin, nx2), numpy.float32)
        ty2 = numpy.array(readfit2dfloats(fin, ny2), numpy.float32)
        cf2 = numpy.array(readfit2dfloats(fin, (nx2 - 4) * (ny2 - 4)), 
                     numpy.float32)
        fin.close()
        # The 3 ,3 is the number of knots
        self.tck1 = (tx1, ty1, cf1, 3, 3)
        self.tck2 = (tx2, ty2, cf2, 3, 3)




class perfect(correctorclass):
    """
    To use on previously corrected when there is no splinefile
    Allows pixel size etc to be set
    """
    splinefile = "NO_CORRECTION_APPLIED"
    xsize = "UNKNOWN"
    ysize = "UNKNOWN"
    def __init__(self):
        correctorclass.__init__(self, None)
    def correct(self, xin, yin):
        """
        Do nothing - just return the same values
        """
        return xin, yin

    def make_pixel_lut(self, dims):
        """
        Generate an x and y image which maps the array indices into
        floating point array indices (to be corrected for pixel size later)

        returns 
        FIXME - check they are the right way around
                add some sort of known splinefile testcase
        """
        # Cache the value in case of multiple calls
        if self.pixel_lut is None:
            x_im = numpy.outer(numpy.arange(dims[0]), numpy.ones(dims[1]))
            y_im = numpy.outer(numpy.ones(dims[0]), numpy.arange(dims[1]))
            self.pixel_lut = x_im, y_im
        return self.pixel_lut



#
#"""
#http://homepages.inf.ed.ac.uk/rbf/CVonline/LOCAL_COPIES/OWENS/LECT5/node5.html
#
#Various interpolation schemes can be used. 
# A common one is bilinear interpolation, given by
#
#v(x,y) = c1x + c2y + c3xy + c4,
#
#where v(x,y) is the grey value at position (x,y).
#Thus we have four coefficients to solve for. We use the known grey values 
# of the 4 pixels
#surrounding the `come from' location to solve for the coefficients.
#
#We need to solve the equation
#
#v1   ( x1 y1 x1y1 1 ) c1
#v2 = ( x2 y2 x2y2 1 ) c2
#v3   ( x3 y3 x3y3 1 ) c3
#v4   ( x4 y4 x4y4 1 ) c4
#
#
#or, in short,
#[V] = [M][C],
#
#which implies
#[C] = [M]-1[V].
#
# This has to be done for every pixel location in the output image and 
# is thus a lot of computation!
# Alternatively one could simply use the integer pixel position closest 
# to the `come from location'.
# This is adequate for most cases.
#
#"""

#def unwarpimage(image, xpositions, ypositions):
#    """
#    xpositions/ypositions are floats giving pixel co-ords of the input image.
#
#    We need the positions of the pixels in the output image,
#     on the input image.
#
#    Hence, for now,
#    """
#    pass





from __future__ import print_function, division

import numpy as np
from ImageD11 import cImageD11
# Estimate a noise level and determine a threshold

def sigma_cut( img, nsigma=3., cycles=3 ):
    """
    Assume the image has been background corrected and 
    follows a Gaussian noise distribution (e.g. CCD)
    """
    imc = np.ascontiguousarray( img, dtype=np.float32 )
    mean, variance = cImageD11.array_mean_var_cut( imgc.ravel(), n=cycles, cut=nsigma, verbose=0 )
    return cuts[-1], means[-1]


def fix_frelon_lines(img, bkg = None, block=256, nsigma=2.):
    """
    Reduce readout artifacts from frelon fast readout
    """
    cImageD11.frelon_lines
    cut, val = sigma_cut( img, nsigma=2 )
    assert img.shape[1]%block == 0
    blocked = np.reshape( img,
                          (img.shape[0]*img.shape[1]//block, block) )
    bgm = blocked<cut
    drift = (blocked*bgm).sum( axis=1 ) / bgm.sum( axis=1 )
    corrected = np.reshape(img, blocked.shape) - drift[:, np.newaxis]
    return np.reshape( corrected, img.shape )


def binary_clean( mask, out = None ):
    """
    Remove the one pixel stuff
    """
    if out is None:
        out = np.zeros(mask.shape, np.int8)
    npx = cImageD11.clean_mask( mask.astype( np.int8 ), out )
    return npx, out
    

def threshold_image(im, nsigma=2, fix_lines=False, mskmem=None, block=256,
                   verbose=0):
    """
    Process one single image (numpy array)
    """
    if fix_lines:
        im = fix_frelon_lines(im, block=block, nsigma=nsigma)
        if verbose: print("fixed lines")
    cut, val = sigma_cut(im, nsigma)
    if verbose:
        print("cut,val: ",cut, val)
    mask = im > cut
    if verbose:
        print("mask raw:",mask.sum())
    npx, clean_mask = binary_clean( mask, out=mskmem )
    i = np.empty( npx, np.uint16)
    j = np.empty( npx, np.uint16)
    tmp = np.empty( im.shape[0], np.int32 )
    ok = cImageD11.mask_to_coo( clean_mask.astype(np.int8), i, j, tmp )
    if verbose:
        print("mask clean",clean_mask.sum())
    assert ok == 0
    print(i.shape,j.shape,clean_mask.sum())
    frm = sparse_frame( i, j, im.shape, itype=np.uint16, 
                       pixels = {'data': im[clean_mask.astype(np.bool)]} )
    return frm

def threshold_series( imseries, nsigma=2, fix_lines=False):
    mskmem = None
    for im in imseries:
        if mskmem is None:
            mskmem = np.zeros( im.shape, np.bool )
        yield threshold_image( im, nsigma=nsigma, 
                              fix_lines = fix_lines, mskmem=mskmem )
        

    
    
        





from __future__ import print_function
from ImageD11 import cImageD11
import fabio
import numpy as np
import scipy.sparse
import unittest
import time

import platform
if platform.system() == "Windows":
    timer = time.clock
else:
    timer = time.time


class test_array_bounds_localmax( unittest.TestCase):
    def setUp(self):
        self.row = np.array([  0,   1,  26,  27, 143, 144, 147, 147,
                               188, 188, 196, 196, 196,
       197, 197, 197, 197, 197, 198, 198, 198, 198, 198, 198, 199, 199,
       199, 200, 200, 201], dtype=np.uint16)
        self.col = np.array([  1,   1, 321, 321, 332, 332, 696, 697,
                               616, 617, 767, 768, 769,
       767, 768, 769, 770, 771,   1,   2, 767, 768, 769, 770, 768, 769,
       770, 772, 773, 773], dtype=np.uint16)
        self.vals = np.array([ 18.757812,  20.311476,  20.125   ,
                               20.809525,  30.141733,
        18.84375 ,  20.023438,  20.023438,  19.328125,  19.328125,
        22.373016,  30.491804,  44.491806,  33.220474, 126.64    ,
       134.64    ,  39.64    ,  22.64    ,  20.882812,  19.882812,
        44.04762 , 109.70248 , 129.70248 ,  32.702477,  43.264   ,
        41.264   ,  33.264   ,  20.593496,  33.5935  ,  35.02362 ],
      dtype=np.float32)
    def test1(self):
        import h5py
        nnz = self.row.shape[0]
        labels = np.zeros( nnz, "i" )
        vmx = np.zeros( nnz, np.float32 )
        imx = np.zeros( nnz, 'i')
        nlabel = cImageD11.sparse_localmaxlabel(
            self.vals, self.row, self.col,
            vmx, imx, labels )
        self.assertTrue( nlabel > 0)
        # This does a double-free on exit

class test_array_stats( unittest.TestCase ):
    def test1( self ):
        ar = np.arange( 6*4, dtype=np.float32 ).reshape( 6,4 )
        mini, maxi, mean, var = cImageD11.array_stats( ar.ravel() )
        t = np.allclose( (mini, maxi, mean, var),
                (ar.min(), ar.max(), ar.mean(), ar.var()) )
        if not t:
            print((mini, maxi, mean, var))
            print((ar.min(), ar.max(), ar.mean(), ar.var()) )
        self.assertTrue(t)
        
    def test2( self ):
        ar = np.ones( 6*4, dtype=np.float32 ).reshape( 6,4 )
        mini, maxi, mean, var = cImageD11.array_stats( ar.ravel() )
        self.assertTrue( np.allclose( 
            (mini, maxi, mean, var),
            (ar.min(), ar.max(), ar.mean(), ar.var()) ))
    def test3( self ):
        ar = np.arange( 2048*2048, dtype=np.float32 ).reshape( 2048, 2048 )
        start = timer() 
        mini, maxi, mean, var = cImageD11.array_stats( ar.ravel() )
        endc = timer()
        check = ar.min(), ar.max(), ar.mean(), ar.var()
        endn = timer()
        self.assertTrue( np.allclose( (mini, maxi, mean, var), check ) )
        numpytime = endn - endc
        ctime = endc - start
        print("array_stats: %.3f ms vs %.3f ms, speedup %.1f"%(
            1e3*ctime, 1e3*numpytime, numpytime/ctime))


class test_array_histogram( unittest.TestCase ):
    def test1(self):
        N = 255
        a = (np.random.random( (2048, 2048))*(N-0.1)).astype(np.float32)
        start = timer()
        hc = np.zeros( N, dtype=np.int32 )
        cImageD11.array_histogram( a.ravel(), 0., N, hc )
        endc= timer()
        hn = np.bincount( a.ravel().astype(int) )
        endn = timer()
        if( hc.shape[0] != hn.shape[0]):
            print("shape error")
            print( a.max())
            print(hc)
            print(hn)
        self.assertTrue( (hc == hn).all() )
        numpytime = endn - endc
        ctime = endc - start
        print("array_histogram: %.3f ms vs %.3f ms, speedup %.1f"%(
            1e3*ctime, 1e3*numpytime, numpytime/ctime))


class test_sparse_is_sorted( unittest.TestCase ):
    def setUp( self ):
        self.i = np.arange(10, dtype = np.uint16 )
        self.j = np.arange(10, dtype = np.uint16 )
    def testOK( self ):
        ret = cImageD11.sparse_is_sorted( self.i, self.j)
        self.assertEqual( ret, 0)
    def testrpt( self ):
        self.i[5]=self.i[4]
        self.j[5]=self.j[4]
        ret = cImageD11.sparse_is_sorted( self.i, self.j)
        self.assertEqual( ret, -5)
    def testbadi( self ):
        self.i[5]=self.i[4]-1
        ret = cImageD11.sparse_is_sorted( self.i, self.j)
        self.assertEqual( ret, 5)
    def testbadj( self ):
        self.i[5]=self.i[4]
        self.j[5]=self.j[4]-1
        ret = cImageD11.sparse_is_sorted( self.i, self.j)
        self.assertEqual( ret, 5)


class test_sparse_connected_pixels( unittest.TestCase ):
    def setUp( self ):
        img ="""
00110100000000000000000000000000000000000000000000000000
00001000000000000000000000000000000000000000000000000000
00000000001111000000111110000000000000000000000000000010
00000000011001000001100011000000000000000000000000010010
11000000000000000000000000000000000000000000000000001111
00000000000000000000000000000000020000000000000000000000
00000000000000000000000000000000020000000000000000000000
00000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000001
00000000000000000001100000000000000000000000000000010001
00000000000000000010011000000000000000000000000000001110
00000000000000000100000100000000000000000000000000000000
00000000000000001000200010000000000000000000000000000000
00000000000000001000000100000000000000000000000000000000
00000000000000000100001000000000000000000000000000000000
00000000000000000011110000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000
00001100000000000000000000000000000000000000000000000000
11111100000000000000000000000000000000000000000000000000
00010000000000000000000100000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000
00000000000000000000000010000000000000000000000100000000
"""
        self.a = np.array([ [int(c) for c in line ] for line in img.split() ],
                          dtype=np.float32)
        self.s = scipy.sparse.coo_matrix( self.a )
        self.threshold=0.5
        if 0:
            print(self.s.data)
            print(self.s.row)
            print(self.s.col)
        
    def testOK( self ):
        #import pdb; pdb.set_trace()
        i = self.s.row.astype( np.uint16 )
        j = self.s.col.astype( np.uint16 )
        v = self.s.data
        self.assertTrue( (self.s.todense() == self.a).all() ) 
        dl = np.zeros( self.a.shape, 'i' )
        nd = cImageD11.connectedpixels(  self.a , dl, self.threshold )
        dstart = timer()
        nd = cImageD11.connectedpixels(  self.a , dl, self.threshold )
        dend = timer()
        sl = np.zeros( v.shape, 'i' )
        ns = cImageD11.sparse_connectedpixels( v, i, j, self.threshold, sl)
        sbegin = timer()
        ns = cImageD11.sparse_connectedpixels( v, i, j, self.threshold, sl)
        send = timer()
        sld = scipy.sparse.coo_matrix( (sl, (i, j)), shape=self.a.shape)
        testcase = (sld.todense() == dl).all()

        densetime = dend - dstart
        sparsetime = send-sbegin
        print("sp_cp: %.3f ms vs %.3f ms, ratio %.1f nnz %d"%(
            1e3*sparsetime, 1e3*densetime, densetime/sparsetime, sld.nnz),
            " t=%.0f"%(self.threshold))
        if ~testcase:
            import pylab as pl
            pl.ioff()
            pl.figure()
            pl.subplot(311)
            pl.title("t=%f"%(self.threshold))
            pl.imshow( dl, interpolation='nearest')
            pl.colorbar()
            pl.subplot(312)
            pl.imshow( sld.todense(), interpolation='nearest')
            pl.colorbar()
            pl.subplot(313)
            pl.imshow( sld.todense() - dl, interpolation='nearest')
            pl.colorbar()
            pl.show()
        self.assertTrue( testcase )

    def testsplat( self ):
        i = self.s.row.astype( np.uint16 )
        j = self.s.col.astype( np.uint16 )
        v = self.s.data
        self.assertTrue( (self.s.todense() == self.a).all() ) 
        ni, nj = self.s.shape
        z = np.empty( (ni+2)*(nj+2), np.int32)
        sl = np.zeros( v.shape, 'i' )
        sls = np.zeros( v.shape, 'i' )
        ns = cImageD11.sparse_connectedpixels( v, i, j, self.threshold, sl)
        s0 = timer()
        ns = cImageD11.sparse_connectedpixels( v, i, j, self.threshold, sl)
        s1 = timer()
        ns = cImageD11.sparse_connectedpixels_splat( v, i, j, self.threshold, sls, z, ni, nj)
        s2 = timer()
        ns = cImageD11.sparse_connectedpixels_splat( v, i, j, self.threshold, sls, z, ni, nj)
        s3 = timer()
        sld = scipy.sparse.coo_matrix( (sl, (i, j)), shape=self.a.shape)
        testcase = (sls == sl).all()
        print("\nsparse_connectedpixels       %.3f ms"%(1e3*(s1-s0)),
              "\nsparse_connectedpixels_splat %.3f ms"%(1e3*(s3-s2)))
        if ~testcase:
            import pylab as pl
            pl.ioff()
            pl.figure()
            pl.subplot(311)
            pl.title("t=%f"%(self.threshold))
            pl.imshow( dl, interpolation='nearest')
            pl.colorbar()
            pl.subplot(312)
            pl.imshow( sld.todense(), interpolation='nearest')
            pl.colorbar()
            pl.subplot(313)
            pl.imshow( sld.todense() - dl, interpolation='nearest')
            pl.colorbar()
            pl.show()
        self.assertTrue( testcase )
        
class test_sparse_connected_pixels_overlapim1(test_sparse_connected_pixels):
    def setUp(self):
        im = fabio.open("testoverlaps0000.edf").data.astype( np.float32 )
        self.a = np.where( im > 1500, im, 0 )
        self.s = scipy.sparse.coo_matrix( self.a )
        self.threshold = 2000.1

class test_sparse_connected_pixels_overlapim2(test_sparse_connected_pixels):
    def setUp(self):
        im = fabio.open("testoverlaps0000.edf").data.astype( np.float32 )
        self.a = np.where( im > 2000.1, im, 0 )
        self.s = scipy.sparse.coo_matrix( self.a )
        self.threshold = 2000.1

class test_sparse_connected_pixels_overlapim3(test_sparse_connected_pixels):
    def setUp(self):
        im = fabio.open("testoverlaps0000.edf").data.astype( np.float32 )
        self.a = np.where( im > 500.1, im, 0 )
        self.s = scipy.sparse.coo_matrix( self.a )
        self.threshold = 500.1

class test_sparse_connected_pixels_overlapim4(test_sparse_connected_pixels):
    def setUp(self):
        im = fabio.open("testoverlaps0000.edf").data.astype( np.float32 )
        self.a = np.where( im > 5000.1, im, 0 )
        self.s = scipy.sparse.coo_matrix( self.a )
        self.threshold = 5000.1
        

if __name__=="__main__":
    unittest.main()


#    cd .. && python setup.py build && cd test && PYTHONPATH=../build/lib.linux-x86_64-2.7 python test_sparse_image.py





from ImageD11.columnfile import columnfile
import sys, numpy as np


cf = columnfile(sys.argv[1])
x = float(sys.argv[2])
y = float(sys.argv[3])
cor = float(sys.argv[4])
tol = float(sys.argv[5])

r = cf.omega * np.pi/ 180.0

xcalc = np.sin(r)*x + np.cos(r)*y + cor

cf.filter( abs (cf.xpos - xcalc) < tol )

cf.writefile( sys.argv[6] )




from __future__ import print_function

from ImageD11 import sym_u
from ImageD11.grain import read_grain_file

c = sym_u.cubic()
c.makegroup()

import glob

fl = glob.glob("x*.ubi")

for f in fl:
    gl = read_grain_file( f )
    x = int(f.split("_")[0][1:])
    y = int(f.split("_")[1].split(".")[0][1:])
    for g in gl:
        ubi = sym_u.find_uniq_u( g.ubi, c )
        g.set_ubi( ubi )
        print(x, y, end=' ')
        for i in range(3):
            for j in range(3):
                print(g.u[i][j], end=' ')
        print()
        
        





# ImageD11_v0.4 Software for beamline ID11
# Copyright (C) 2005  Jon Wright
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

try:
    from tkMessageBox import showinfo, askyesno
except:
    # python3?
    from tkinter.messagebox import showinfo, askyesno

from . import twodplot


class guipeaksearcher:
    """
    Tkinter wrapper for ImageD11 guipeakmerge object

    Note that peaksearching is done via the command line

    All communication should be via parent guicommander object
    """
    def __init__(self,parent,quiet="No"):
        """
        Parent is a hook to features of the parent gui
        sets up peakmerging menu items + message for searching
        """
        self.parent=parent
        self.quiet=quiet
        #print "I am in quiet mode",quiet
        self.menuitems = ( "PeakSearching", 4,
                           [ ( "Search raw images" , 0, self.searchraw ),
                             ( "Read pks file", 0, self.readpeaks ),
                             ( "Harvest peaks", 0, self.harvestpeaks),
                             ( "Merge peaks",  0, self.mergepeaks ),
                             ( "Filter peaks", 0, self.filter     ),
                             ( "Save good peaks",2, self.savepeaks) ] )

    def searchraw(self):
        """ Explains to user about the command line script """
        showinfo("Sorry","""Not implemented for gui, please use the jonpeaksearch script on amber/crunch for now""")

    def readpeaks(self,filename=None):
        """
        Runs peakmerger.readpeaks
        gets names for first and last image for plot title
        gets omega angles and image numbers for plot
        plots image number versus omega angle to see if you did a scan
        """
        if filename is None:
            filename = self.parent.opener.show(title="Peaksearch results file")
        import os
        if os.path.isfile(filename):
            # apply this to peakmerge
            self.parent.guicommander.execute("peakmerger","readpeaks",filename)
        else:
            showinfo("Sorry, bad filename %s"%(filename))
        imagenumbers = self.parent.guicommander.getdata("peakmerger","imagenumbers")
        omegas = self.parent.guicommander.getdata("peakmerger","omegas")
        images = self.parent.guicommander.getdata("peakmerger","images")
        first = images[0].name
        last = images[-1].name
        self.parent.twodplotter.adddata(
              ("number versus omega",
                 twodplot.data(
                  imagenumbers,omegas,
                  {"xlabel" : "imagenumber",
                   "ylabel" : "Omega",
                   "title"  : first+"..."+last }
                  ) # data
                 )# tuple to plotter
              ) # call

    def harvestpeaks(self):
        """
        gets range from 2d plot (image numbers and omegas)
        calls peakmerger.harvestpeaks(image_number_range,omega_range)
        """
        # Now we need to select the range of peaks to use
        if self.quiet=="No":
            ans = askyesno("Have you selected a sensible range of images?","""
   Use the mouse to select the range of image numbers and omega angles that
   you want to use from the graph on the screen.

   If all is OK, then say yes now and we will try to harvest the peaks.
   Otherwise, say no, select the right range and come back "harvestpeaks" again
   """       )
            # FIXME? ans is ignored anyway
        # We now have the ranges in imagenumber and omega from
        # the graph
        numlim=self.parent.twodplotter.a.get_xlim()
        omlim=self.parent.twodplotter.a.get_ylim()
        self.parent.guicommander.execute("peakmerger","harvestpeaks",numlim=numlim,omlim=omlim)
        if self.quiet=="No":
            npks = len(self.parent.guicommander.getdata("peakmerger","allpeaks") )
            showinfo("Harvested peaks","You have a total of %d peaks,"%(npks)+
               "no peaks have been merged")

    def mergepeaks(self):
        """ calls peakmerger.mergepeaks and reports number of peaks to user """
        self.parent.guicommander.execute("peakmerger","mergepeaks")
        nmerged = len(self.parent.guicommander.getdata("peakmerger","merged"))
        if self.quiet=="No":
            showinfo("Finished merging peaks","You have a total of "+str(nmerged)+" after merging")
        return

    def filter(self):
        """ calls peakmerger.filter (does very little)
        plots x and y final peak positions

        TODO implement filters!!!
        """
        self.parent.guicommander.execute("peakmerger","filter")
        peaks = self.parent.guicommander.getdata("peakmerger","finalpeaks")
        if self.quiet=="No":
            self.parent.twodplotter.hideall() # get rid of image number versus omega plot
            self.parent.twodplotter.adddata(
                  ( "Filtered peaks",
                     twodplot.data(
                        peaks[0,:],
                        peaks[1,:],
                        {"xlabel" : "x",
                         "ylabel" : "y",
                         "title"  : "Peak positions on detector"}
                         ) ) )
        return "nothing"
        # Need to filter based on x,y
        # also based on intensity
        # also based on shape

    def savepeaks(self,filename=None):
        """ see peakmerger.savepeaks """
        if filename==None:
            filename=self.parent.saver.show(title="Filtered peak positions")
        self.parent.guicommander.execute("peakmerger","savepeaks",filename)




from __future__ import print_function


#
from ImageD11 import transform, transformer, indexing, parameters
from ImageD11.closest import score
import sys

    


def gridgrains( ul, flt, pars, 
                minx=-750, maxx=750, stepx=25,
                miny=-750, maxy=750, stepy=25,
                tol = 0.1,
                ) :
    trn = transformer.transformer()
    trn.loadfiltered( flt )
    trn.parameterobj = parameters.parameters( **pars )

    peaks = [trn.getcolumn(trn.xname),
             trn.getcolumn(trn.yname) ]
    peaks_xyz =  transform.compute_xyz_lab( peaks,
                       **trn.parameterobj.get_parameters() )
    
    omega = trn.getcolumn( trn.omeganame )
    trn.updateparameters()
    tx = minx - stepx
    n = 0
    while tx <= maxx:
         tx = tx + stepx
         ty = miny - stepy
         while ty <= maxy:
             ty = ty + stepy      
             trn.parameterobj.set_parameters(  {
                 't_x': tx,
                 't_y': ty })
             pars = trn.parameterobj.get_parameters()
             tth, eta = transform.compute_tth_eta_from_xyz(
                 peaks_xyz,
                 omega,
                 **pars)
             if 'omegasign' in pars:
                 om_sgn = float(pars["omegasign"])
             else:
                om_sgn =  1.0
                
             gv = transform.compute_g_vectors(
                 tth, eta, omega*om_sgn, **pars)
             print(tx, ty, end=' ')
             for ubi in ul:
                 ns = score( ubi, gv.T, tol )
                 print(ns, end=' ')
                 n += ns
             print()
    return n


if __name__ == "__main__":
    flt = sys.argv[1]
    par = sys.argv[2]
    grains = sys.argv[3]

    ul = indexing.readubis( grains )

    po = parameters.parameters()
    po.loadparameters( par )
    pars = po.get_parameters()

    dcen =float(  pars["distance"] )
    zcen =float(  pars["z_center"] )
    ycen =float(  pars["y_center"] )
    f = open("parmap","a")
    for dist in [dcen*0.99, dcen, dcen*1.01]:
        pars["distance"]=dist
        for yc in [ ycen-1 ,ycen ,ycen+1 ]:
            pars["y_center"] = yc
            for zc in [zcen-1, zcen, zcen +1 ]:
                pars["z_center"] = zc
                print("\n\n# %f %f %f"%(dist, yc, zc))
                npk = gridgrains( ul, flt, pars )
                f.write("%f %f %f %d\n"%( dist, yc, zc, npk))
                f.flush()
                





# Create objects to manipulate - they hold your data
#
from ImageD11 import indexing
myindexer = indexing.indexer()

#
# Your work starts here:
#
myindexer.readgvfile(  'Al1000/Al1000.gve' )
myindexer.updateparameters( )
myindexer.parameterobj.set_parameters(  {
    'minpks': 100,
    'max_grains': 1001,
    'ring_1': 5,
    'ring_2': 5,
    'cosine_tol': -0.001,
    } )
myindexer.loadpars( )
myindexer.assigntorings( )
myindexer.find( )
myindexer.scorethem( )
myindexer.saveubis(  'allgrid.map' )



from __future__ import print_function
from fabio import tifimage
from ImageD11 import columnfile
import numpy as np
import os, unittest, sys


class test_tifs(unittest.TestCase):

    FIRST=6
    LAST =16
    NPK  =0 

    CMD  = sys.executable + " " + os.path.join( 
            "..","..", "scripts", "peaksearch.py " )
    
    def setUp(self):
        blank = np.zeros((24,24),np.uint16)
        peak   = blank.copy()
        peak[ 12:14, 15:17 ] = 42
        
        # put a peak on every other image
        for i in range(self.FIRST, self.LAST+1):
            if i%2 == 0:
                obj = tifimage.tifimage( peak, { } )
                self.NPK += 1
                print( i, self.NPK)
            else:
                obj = tifimage.tifimage( blank, { } )
            obj.write( 'tiftest%04d.tif'%(i))
        
       
    def tearDown(self):
        return
        # we'll leave the files around for now

        for i in range(self.FIRST, self.LAST+1):
            os.remove( 'tiftest%04d.tif'%(i) )
        


    def testpeaksearch(self):
        os.system(self.CMD + " -n tiftest " + \
                  " -t 5 -F .tif -S 0.3 " + \
                  " -T 11 -p Y --OmegaOverRide " +\
                  " -f %d -l %d"%(self.FIRST, self.LAST))
        results = columnfile.columnfile( "peaks_t5.flt" )
        self.assertEqual( results.nrows, self.NPK)
        if os.path.exists( 'peaks_t6.flt' ):
            self.assertEqual( open( 'peaks_t5.flt').read().rstrip(),
                              open( 'peaks_t6.flt').read().rstrip() )

    def testpeaksearch_nooverride(self):
        os.system(self.CMD + " -n tiftest " + \
                  " -t 6 -F .tif -S 0.3 " + \
                  " -T 11 -p Y " +\
                  " -f %d -l %d"%(self.FIRST, self.LAST))
        results = columnfile.columnfile( "peaks_t6.flt" )
        self.assertEqual( results.nrows, self.NPK)
        if os.path.exists( 'peaks_t5.flt' ):
            self.assertEqual( open( 'peaks_t5.flt').read().rstrip(),
                              open( 'peaks_t6.flt').read().rstrip() )
        
    def testpeaksearch_singlethread(self):
        os.system(self.CMD + " -n tiftest " + \
                  " -t 7 -F .tif -S 0.3 " + \
                  " -T 11 -p Y --singleThread " +\
                  " -f %d -l %d"%(self.FIRST, self.LAST))
        results = columnfile.columnfile( "peaks_t7.flt" )
        self.assertEqual( results.nrows, self.NPK)
        if os.path.exists( 'peaks_t6.flt' ):
            self.assertEqual( open( 'peaks_t7.flt').read().rstrip(),
                              open( 'peaks_t6.flt').read().rstrip() )
        
        
if __name__ == '__main__':
    unittest.main()
    
   
    
    




#!/usr/bin/env python
from __future__ import print_function

from ImageD11.columnfile  import columnfile
from ImageD11.grain import read_grain_file, write_grain_file
import sys, os, multiprocessing, platform

try:
    c = columnfile( sys.argv[1] )
    g = read_grain_file( sys.argv[2] )
    parfile = sys.argv[3]
    cmds = []
except:
    print( "Usage: %s colfile.flt.new grains.map parameters.par  --omega_slop=1 etc"%(sys.argv[0]))
    sys.exit()
    

for i in range(len(g)):
    #g[i].translation[2] = 0.0
    write_grain_file("%d.ubi"%(i),[g[i]])
    d = c.copy()
    d.filter( d.labels == i )
    d.writefile("%d.flt"%(i))
    cmd = "%s %s/fitgrain.py"%( sys.executable, os.path.split(__file__)[0] )
    cmd += " -p %s -u %d.ubi -U %d.ubi -P %d.par -f %d.flt -x t_z"%(
        parfile,i,i,i,i)
    for extra_arg in sys.argv[4:]:
        cmd += " "+extra_arg
    cmds.append( cmd )

if platform.system() != "Windows":
    p = multiprocessing.Pool( multiprocessing.cpu_count() )
    p.map( os.system, cmds )
else:
    for cmd in cmds:
        print(cmd)
        os.system(cmd)





from __future__ import print_function, division
from ImageD11 import cImageD11, indexing


try:
    from time import perf_counter_ns
except:
    from time import time
    def perf_counter_ns():
        return time()*1e9

import numpy as np
i = indexing.indexer()
i.readgvfile("eu3.gve")

gve = i.gv.copy().astype(np.float)
#gve = np.random.random( size=(100000,3) )-0.5
rgba = np.zeros( (512, 512, 4), np.uint8 )
u = np.eye(3, dtype=np.float).ravel()
t0 = perf_counter_ns()
cImageD11.splat( rgba, gve, u, 1 )
t1 = perf_counter_ns()
for j in range(5):
    cts = []
    for i in range(10):
        t2 = perf_counter_ns()
        cImageD11.splat( rgba, gve, u, j )
        t3 = perf_counter_ns()
        cts.append(t3-t2)
    print("Ran that for npx=",j,"s, fps:", 1e9/np.mean(cts),1e9/np.max(cts),1e9/np.min(cts))
import pylab as pl
pl.imshow(rgba)
pl.show()




from __future__ import print_function
import glob, sys, os

import numpy as np

from ImageD11 import indexing, sym_u, closest


# open up a set of ubi files


ubifl = glob.glob(sys.argv[1]+"/*.ubi")

ubi_all = {}




#print "Make cubic point group"
c = sym_u.cubic()
c.makegroup()

for fname in ubifl:
    #print fname

    # Flip to standard setting
    ubi_all[fname] = [ sym_u.find_uniq_u( ubi, c ) for ubi in
                       indexing.readubis(fname) ]


# Now try to make these into a unique set.
#
# For each ubi:
#     compute the reciprocal peak positions of central lattice
#     see if these match a previous one.

pks = [

        ( 5, 3,-1), ( 5,-3, 1), (-5, 3, 1),
        (-5,-3, 1), (-5, 3,-1), ( 5,-3,-1),
        (-5,-3,-1), ( 5, 3, 1),
        ( 5, 1,-3), ( 5,-1, 3), (-5, 1, 3),
        (-5,-1, 3), (-5, 1,-3), ( 5,-1,-3),
        (-5,-1,-3), ( 5, 1, 3),
        ( 1, 3,-5), ( 1,-3, 5), (-1, 3, 5),
        (-1,-3, 5), (-1, 3,-5), ( 1,-3,-5),
        (-1,-3,-5), ( 1, 3, 5),
        ( 3, 5,-1), ( 3,-5, 1), (-3, 5, 1),
        (-3,-5, 1), (-3, 5,-1), ( 3,-5,-1),
        (-3,-5,-1), ( 3, 5, 1),
        ( 1, 5,-3), ( 1,-5, 3), (-1, 5, 3),
        (-1,-5, 3), (-1, 5,-3), ( 1,-5,-3),
        (-1,-5,-3), ( 1, 5, 3),
        ( 3, 1,-5), ( 3,-1, 5), (-3, 1, 5),
        (-3,-1, 5), (-3, 1,-5), ( 3,-1,-5),
        (-3,-1,-5), ( 3, 1, 5),

        ]
#print len(pks)
hkl = np.array(pks,np.float).T

uniq_ubis = []
names = list(ubi_all.keys())

tol = 0.25
for name in names:
    this_list = ubi_all[name]
    for ubi, i in zip(this_list,list(range(len(this_list)))):
        gv = np.dot(np.linalg.inv(ubi), hkl)
        seen = 0
        for j in range(len(uniq_ubis)):
            u = uniq_ubis[j][0]
            npk = closest.score(u,gv.T,tol)
            if npk == len(pks):
                # seen you before
                uniq_ubis[j][1].append((name, i))
                uniq_ubis[j][2].append(ubi)
                uniq_ubis[j][0] = np.add.reduce( uniq_ubis[j][2] )/ \
                                            len(uniq_ubis[j][2])
                seen += 1
            #if npk > 50 and npk < len(pks):
            #    print "close...",npk,
        if seen == 0:
            uniq_ubis.append([ubi,[(name,i)], [ubi] ])
        if seen > 1:
            print("uniq seen more than once",ubi,i,name)

print("# Found",len(uniq_ubis),"unique ubi matrices")

dsu = [ (len(uo[1]),(uo[1],uo[0])) for uo in uniq_ubis]
dsu.sort()

for l,uo in dsu[::-1]:
    ubi = uo[1]
    # print "\n# found",len(uo[0]),"times"
    # Find average ubi matrix for each time...
    usum = np.zeros((3,3),np.float)
    j = 0
    for name, i in uo[0]:
        usum = usum + ubi_all[name][i]
        j += 1
    u = usum/j
    print("%f %f %f\n"  %(u[0,0],u[0,1],u[0,2]), end=' ')
    print("%f %f %f\n"  %(u[1,0],u[1,1],u[1,2]), end=' ')
    print("%f %f %f\n\n"%(u[2,0],u[2,1],u[2,2]), end=' ')

j = 0
for l, uo in dsu[::-1]:
    print("# ", uo[1][0])
    print("# ", uo[1][1])
    print("# ", uo[1][2])
    j += 1
    for name, i in uo[0]:
        n = os.path.split(name)[1]
        x = int(n.split("_")[0][1:])
        y = int(n.split("_")[1].split(".")[0][1:])
        print("%d %d %d"%(x,y,j))





from numpy import dot, round_, array, float, allclose, asarray, fabs,\
        argmax, sqrt, argsort, take, sum, where
from numpy.linalg import inv

DEBUG = False

# Some sort of round off
MIN_VEC2 = 1e-9*1e-9


def fparl(x, y):
    """ fraction of x parallel to y """
    ly2 = dot(y ,y)
    if ly2 > 1e-9:
        return dot(x, y) / ly2
    return 0

def mod(x,y):
    """ 
    Returns the part of x with integer multiples of y removed
    assert that ||mod(x,y)|| <= ||x|| for all y 
    """
    if __debug__: 
        b4  = dot(x,x)
    n   = round_(fparl(x,y))
    ret = x - n * y
    if __debug__:
        af  = dot(ret,ret)
        assert b4 >= af , "Bad mod "+str(x) + " " + str(y)
    return ret


def sortvec_len( vl ):
    """
    Sorts according to length (d-s-u)
    """
    ul = [ (dot(v,v),tuple(v)) for v in vl ]
    # Choose x,y,z
    ul.sort()
    return [ asarray(v[1]) for v in ul ]

def sortvec_xyz( vl ):
    """
    For each choose between x, -x
    Sorts according to x then y then z components
    """
    ul = []
    # Choose x/-x
    for v in vl:
        # Wonder how to do this with numpy?
        ul.append( max( tuple(v),tuple(-asarray(v)) ) )
    # Choose x,y,z
    ul.sort()
    return [ asarray(v) for v in ul[::-1] ]



class lattice(object):
    """
    Represents a 3D crystal lattice built from 3 vectors
    """

    def __init__(self, v1, v2, v3, space='reciprocal', min_vec2=MIN_VEC2):
        """ Make the lattice 
        Currently v1, v2, v3 are vectors - which gives 3D
        space [ 'real' | 'reciprocal' ]
        ... means they are real space lattice vectors or measured scattering
            vectors from crystallography
        It will attempt to find a primitive basis from the supplied vectors
        """
        # 3 supplied vectors (these are lattice vectors)
        vl=[v1, v2, v3] 
        # 9 difference vectors (these are also lattice vectors)
        # Assuming : We can always add or subtract integer amounts of any
        #            lattice vector to get another lattice vector.
        #            We look for the 3 shortest which are non-coplanar
        #
        # We could do this recursively adding and subtracting
        # ... We *need* to do it only until the solution is stable 
        # one would hope this is enough, but more testcases should be added?
        # ... currently the algorith is vague
        dsu = sortvec_len( vl + [  vi - vj for vi in vl for vj in vl ] +
                                [  vi + vj for vi in vl for vj in vl ]  )
        # Sort by length
        # Take shortest as first vector
        vl = []
        for i in range(len(dsu)):
            t = dsu[i]
            for v in vl: # remove previous vectors
                t = mod( t, v )
            if len(vl) == 2: # And try to catch hexagonal planes too
                t = mod( t, vl[0] + vl[1] )
                t = mod( t, vl[0] - vl[1] )
            if dot(t,t) < min_vec2: # Skip if zeroed out
                continue
            vl.append( t )
            if len(vl) == 3: break
            # Remove this from future vectors ???
            dsu = [ mod( d , t ) for d in dsu]
        assert len(vl) == 3, "Have not got 3 vectors"
        # print "vl:",vl
        vl = sortvec_xyz(vl)
        # print "vl",vl
        if space == 'reciprocal':  
            self.v  = array(vl).T   # We got column vectors
            self.vi = inv(self.v)
        elif space == 'real':
            self.vi = array(vl)     # We got row vectors
            self.v  = inv(self.vi) 
        else:
            raise Exception("Space must be real or reciprocal "+str(space))
    def ijks(self, u):
        """
        Give back the nearest lattice point indices
        ... only present for symmetry with hkls
        """
        return round_(dot(self.v, u))
    def rem_ijks(self, u):
        """
        Remainder of u (distance from a lattice point)
        """
        return g - dot(self.vi, self.hkls(g))
    def hkls(self, g):
        """ 
        Give back the hkl indices 
        ... these are the nearest lattice point
        """
        return round_(dot(self.vi, asarray(g).T)).T
    def rem_hkls(self, g):
        """ 
        Remainder of x after removing closest lattice point in reciprocal space
        """
        h = dot(self.v, self.hkls(g).T).T
        print 'h',h
        assert g.shape == h.shape, "bug!"
        return g - h
    def withvec(self, x, space="reciprocal"):
        """ 
        Try to fit x into the lattice 
        Make the remainder together with current vectors
        Index it as hkl indices
        whichever vector has the biggest projection is replaced
        remake the lattice with these 3 vectors
        """
        x = asarray(x)
        print "withvec",x
        print 'v',self.v
        print 'vi',self.vi
        if space == 'reciprocal':
            r = self.rem_hkls(x)
            print "r",r
            vd = argmax(fabs(dot(self.vi, x)))
            print "vd", vd
            v = list(self.v.T)  # reciprocal space is columns
            v[vd] = r
        if space == 'real':
            r = self.rem_ijks(x)
            print "r",r
            vd = argmax(fabs(dot(self.vi, x)))
            print "vd",vd
            v = list(self.vi)
            v[vd] = r
        print "v",v
        return lattice( v[0], v[1], v[2] , space=space )
    def score_recip(self, g, tol=0.1):
        """
        How many peaks have rem less than tol
        """
        ga = asarray(g)
        assert ga.shape[1] == 3, "gv shape"
        #print ga.shape
        r = self.rem_hkls(ga)
        assert ga.shape == r.shape
        r2 = sum( r*r, axis=1 )
        s = sum( where( r2 < tol * tol, 1, 0) )
        return s

def get_eu_gv():
    from ImageD11.indexing import indexer, ubitocellpars
    o = indexer()
    o.readgvfile("eu3.gve" , quiet = True)
    return o.gv

def test_fft():
    gv = get_eu_gv()
    from ImageD11.fft_index_refac import grid
    from ImageD11.indexing import ubitocellpars, write_ubi_file,\
        refine
    g = grid( np = 128,
              mr = 1.0,
              nsig = 20,
              minlen = 3. )
    g.gv_to_grid_new(gv)
    g.fft()
    g.props()
    g.peaksearch(open("eu.patterson_pks","w"))
    g.read_peaks("eu.patterson_pks")
    vecs = g.UBIALL
    order = argsort( g.colfile.sum_intensity )[::-1]
    min_pks = 300
    try:
        for i in order:
            print vecs[i], g.colfile.sum_intensity[i]
            for j in order[i:]:
                for k in order[j:]:
                    print i,j,k
                    try:
                        l = lattice( vecs[i], vecs[j], vecs[k],
                                     space='real')
                        t = 0.1
                        ref = refine( l.vi, gv, t)
                        l = lattice( ref[0], ref[1], ref[2], space='real')
                        print "UBI:"
                        print l.vi
                        print ubitocellpars(l.vi)
                        s = l.score_recip( gv , t )
                        print "Indexes",s ,"at tolerance",t
                    
                    except:
                        import traceback
                        print traceback.print_exc()
                        raise Exception("error")
                    if s > min_pks:
                        write_ubi_file( "trial.ubi", [ l.vi ] )
                        raise Exception("Really bad control structure")
    except:
        import traceback
        print traceback.print_exc()
        print "Got something"

def test_eu():
    #Conventional cell gives:
    #0 = (0, 1,  1)
    #1 = (0, 1, -1)
    #6 = (1, 0, -1)
    gv = get_eu_gv()
    v1 = gv[0]
    v2 = gv[1]
    v3 = gv[6]
    l = lattice ( v1, v2, v3)
    print l.v, l.vi
    esum = 0.0
    for v in gv:
        print ("%8.5f "*3+"%4f "*3)%tuple( list(v)+list( 
                dot(l.vi,v.T))),
        err   = l.rem_hkls(v)
        esum += sqrt(dot(err,err))
        print "%.5f"%(sqrt(dot(err,err)))
    # print esum / len(o.gv)
    assert esum / len(gv) < 0.0102, "Did not fit"

def test2():
    """ Adding a fourth vector """
    o = lattice( array( [ 1, 0, 0] , float) ,
                 array( [ 0, 1, 0] , float) ,
                 array( [ 0, 0, 2] , float) )
    # ... how to do it?
    r  = o.rem_hkls( array( [ 0, 0, 1] , float) ) 
    assert ( r == array( [ 0, 0, 1] , float) ).all()
    o = o.withvec( r )
    assert allclose(o.v[0], array( [ 1, 0, 0] , float) )
    assert allclose(o.v[1], array( [ 0, 1, 0] , float) )
    assert allclose(o.v[2], array( [ 0, 0, 1] , float) )
    
def test1():
    """ Make a lattice from 3 vectors"""
    o = lattice( array( [ 1, 0, 0] , float) ,
                 array( [ 2, 1, 0] , float) ,
                 array( [ 3, 4, 1] , float) )
    assert allclose(o.v[:,0], array( [ 1, 0, 0] , float) )
    assert allclose(o.v[:,1], array( [ 0, 1, 0] , float) )
    assert allclose(o.v[:,2], array( [ 0, 0, 1] , float) )
    """ Assign hkl indices in a strange way (no matrix inversing) """
    for x in [ [ 1,3,4 ], [ -9, 10, 13] ]:
        h = o.hkls(x)
        assert  (asarray(h) == asarray(x)).all(), x
    """ 3 more difficult vectors """
    o = lattice( array( [ 1, 0, 1] , float) ,
                 array( [ 0, 1, 0] , float) ,
                 array( [ 0, 0, 1] , float) )
    assert allclose(o.v[:,0], array( [ 1, 0, 0] , float) )
    assert allclose(o.v[:,1], array( [ 0, 1, 0] , float) )
    assert allclose(o.v[:,2], array( [ 0, 0, 1] , float) )

    """ 3 really hard vectors """
    global DEBUG
    DEBUG = True
    o = lattice( array( [ 991, 990, 990] , float) ,
                 array( [ 990, 991, 990] , float) ,
                 array( [ 990, 990, 991] , float) )
    print "o.v",o.v
    print o.hkls( [990,990,990] ) 
    o = o.withvec( [990,990,990] ) # Remove long
    assert allclose(o.v[:,0], array( [ 1, 0, 0] , float) )
    assert allclose(o.v[:,1], array( [ 0, 1, 0] , float) )
    assert allclose(o.v[:,2], array( [ 0, 0, 1] , float) )
    DEBUG = False

        

if __name__=="__main__":
    import time
    start = time.time()
    test1()
    test2()
    test_eu()
    test_fft()
    print time.time()-start
    




from __future__ import print_function, division

header = []
code   = []

header.append("""
#ifndef _write_check_h
#define _write_check_h
void readcpuid( unsigned int leaf, unsigned int subleaf, unsigned int idBits[4] );
""")

code.append("""
/** 
 * This code was auto-generated by write_check.py. Do not edit here!
 * 
 * Utilities for checking the CPU instruction set
 *
 * https://en.wikipedia.org/wiki/CPUID
 *
 */

#include <stdio.h>

#ifdef _MSC_VER
    #include <intrin.h>
#else
    #include <cpuid.h>
#endif
#include "cImageD11.h"
#include "check_cpu_auto.h"

/* DO NOT Use static globals here 
 * Results of calls to cpuid 
 * Stuff to look for : EAX call, A|B|C|D, bit position 
 *  0  : {B,D,C} = name, A = maxcall
 *  1  :  4A, 5B, 6C, 7D
 *      A = Processor information
 *      B = Additional information
 *      C = Features
 *      D = Features
 *  2  : Cache and TLB Descriptor information
 *  3  : Processor Serial Number
 * EAX=4 and EAX=Bh: Intel thread/core and cache topology
 * EAX=7, ECX=0: Extended Features
 * EAX=80000000h: Get Highest Extended Function Implemented : 
 *                The highest calling parameter is returned in EAX.
 * EAX=80000001h: Extended Processor Info and Feature Bits
 * EAX=80000002h,80000003h,80000004h: Processor Brand String
 * EAX=80000005h: L1 Cache and TLB Identifiers
 * EAX=80000006h: Extended L2 Cache Features
 * EAX=80000007h: Advanced Power Management Information
 * EAX=80000008h: Virtual and Physical address Sizes
 * EAX=8FFFFFFFh: AMD Easter Egg
 */


/**
 * Calls cpuid and stores results of eax,ebx,ecx,edx in idBits[op*4:op*4+4]
 * __cpuid_count needed for > XXX
 */
void readcpuid( unsigned int leaf, unsigned int subleaf, unsigned int idBits[4] ) {
  int j;
  // MSVC and gcc both provide a __cpuid function
  for(j=0;j<4;j++) idBits[j]=0;
  #if defined(__GNUC__)
   __cpuid_count( leaf, subleaf, idBits[0], idBits[1], idBits[2], idBits[3]);
  #elif defined(_WIN32)
  __cpuidex( &idBits[0], leaf, subleaf );
  #endif
}


/**
 * Retrieve the processor name.
 * \param name Preallocated string containing at least room for 13 characters. Will
 *             contain the name of the processor.
 */
void cpuidProcessorName( char* name ){
  unsigned int idBits[4];
  readcpuid( 0, 0, idBits );
  name[0]  = (idBits[1]    ) & 0xFF;
  name[1]  = (idBits[1]>> 8) & 0xFF;
  name[2]  = (idBits[1]>>16) & 0xFF;
  name[3]  = (idBits[1]>>24) & 0xFF;
  name[4]  = (idBits[3]    ) & 0xFF;
  name[5]  = (idBits[3]>> 8) & 0xFF;
  name[6]  = (idBits[3]>>16) & 0xFF;
  name[7]  = (idBits[3]>>24) & 0xFF;
  name[8]  = (idBits[2]    ) & 0xFF;
  name[9]  = (idBits[2]>> 8) & 0xFF;
  name[10] = (idBits[2]>>16) & 0xFF;
  name[11] = (idBits[2]>>24) & 0xFF;
  name[12] = 0;
}



""")

for name, byte, bit in [
    ("SSE", 7, 25),
    ("SSE2", 7, 26),
    ("SSE3", 6, 0),
    ("FMA3",6,12),
    ("SSE41",6,19),
    ("SSE42",6,20),
    ("MOVBE",6,22),
    ("XSAVE", 6, 26),
    ("OSXSAVE",6,27),
    ("AVX",6,28),
    ("AVX2", 29,5),
    ("AVX512F",29,16),
    ]:
    func = "int flag_%s(){\n"%(name) + \
           " unsigned int idBits[4];\n" + \
           " readcpuid( %d , 0, idBits );\n"%(byte//4) + \
           " return (idBits[%d] & ( 1 << %d))>0;\n}\n"%(byte%4,bit)
    code.append(func)
    header.append("int flag_%s(void);\n"%(name))

header.append("int i_have_SSE2(void);\n")
header.append("int i_have_SSE42(void);\n")
header.append("int i_have_AVX(void);\n")
header.append("int i_have_AVX2(void);\n")
header.append("int i_have_AVX512F(void);\n")

code.append( """
int i_have_SSE2(){
  return (flag_SSE2()>0);
}
int i_have_SSE42(){
  return (flag_SSE42()>0);
}
int i_have_AVX(){
    return (flag_XSAVE()>0)&(flag_OSXSAVE()>0)&(flag_AVX()>0);
}
int i_have_AVX2(){
    return (flag_AVX()>0)&(flag_MOVBE()>0)&(flag_FMA3()>0)&(flag_AVX2()>0);
}
int i_have_AVX512F(){
    return (flag_OSXSAVE()>0)&(flag_AVX512F()>0);
}

""")


header.append("#endif\n")


def main():
    with open("check_cpu_auto.c", "w") as f:
        f.write("".join( code ) )
    with open("check_cpu_auto.h", "w") as h:
        h.write("".join( header ) )
        

if __name__=="__main__":
    main()



from ImageD11.indexing import readubis
from ImageD11.transform import uncompute_g_vectors
from ImageD11 import unitcell
import numpy as np

ul = readubis("test.ubi")
ui = [ np.linalg.inv(ubi) for ubi in ul ]

r = range(-4,5)
hkls = np.array([ (h,k,l) for h in r for k in r for l in r ])


gcalc = [ np.dot( ub, hkls.T).T for ub in ui]

# 30 keV
energy = 30
wvln = 12.3985 / energy
ng = 0
for g, ubi in zip(gcalc, ul):
    ng +=1
    print "# Grain",ng, "\n# Energy",energy,"keV, wavelength %.5f"%(wvln)
    tth, eta, omega = uncompute_g_vectors( g.T, wvln )
    order = np.argsort(tth)
    # hkls = np.dot( ubi, g.T).T
    print "#   h    k    l   tth     eta    omega"
    for j in range(len(order)):
        i = order[j]
        for s in (0,1):
            h,k,l = [int(v) for v in hkls[i]]
            if tth[i]>0  and tth[i] < 20 and \
                    abs(eta[s][i])< 45 and not unitcell.F(h,k,l):

                
                print (" %4d"*3 + " %7.2f"*3) % tuple([h,k,l,
                        tth[i],
                        eta[s][i], omega[s][i]])
    print



#!/usr/bin/env python
from __future__ import print_function


# ImageD11_v1.0 Software for beamline ID11
# Copyright (C) 2005-2007  Jon Wright
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


"""
Graphical user interface for ImageD11

Uses Tkinter (comes with python, normally)
Also depends eventually on matplotlib (publication quality 2d plotting)
and on OpenGl for plotting things in 3D (reciprocal space)

Hopefully the gui and the gruntwork can be completely separated, eventually.
"""

try:
    from Tkinter import *
    import tkFileDialog as filedialog
    from tkMessageBox import showinfo
    from ScrolledText import ScrolledText
except:
    from tkinter import *
    import tkinter.filedialog as filedialog
    from tkinter.messagebox import showinfo
    from tkinter.scrolledtext import ScrolledText

import logging
import sys
import os

# GuiMaker is for building up the windows etc

from ImageD11.tkGui.guimaker import GuiMaker
from ImageD11.tkGui import twodplot,  guipeaksearch, guitransformer, guiindexer, guisolver
from ImageD11 import __version__, guicommand
from ImageD11.license import license

# This does not do anything unless you call it as a program:

if __name__ == "__main__":
    # get the output!
    # Set up the logging stuff
    console = logging.StreamHandler(sys.stdout)
    # set a format which is simpler for console use
    formatter = logging.Formatter('%(levelname)-8s : %(message)s')
    # tell the handler to use this format
    console.setFormatter(formatter)
    # add the handler to the root logger
    console.setLevel(logging.DEBUG)
    root = logging.getLogger('')
    root.addHandler(console)
    root.setLevel(logging.DEBUG)  # sh

    # Help message - TODO - proper online help
    def help():
        hlp = """Try pointing your web browser at:
   http://fable.sourceforge.net/index.php/ImageD11
   See also the html documentation for programmers, somewhere like:
   file:///c:/python24/Lib/site-packages/ImageD11/doc/ImageD11.html
"""
        showinfo("Help", "Please try harder\n"+hlp)

    # Credits message
    ImageD11credits = """Thanks to:

                        All of the Fable team, which includes at least:
                           Andy Gotz, Gavin Vaughan, Henning O. Sorensen,
                           Soren Schmidt, Henning Poulsen, Larry Margulies
                           Erik Knudsen,
                        ...and others who should remind me to mention them

                        Tine Knudsen who bravely helped commission the
                        introduction of a second rotation axis (wedge).

                        Benoit Mallard for his assistance with some extreme
                        programming to debug the transformation module.

                        Younes ElHachi for adding the eps_sig calculations.

                        John Hunter for the matplotlib plotting.

                        All of the pyopengl, Numeric, numpy and python teams

                        Anyone who tests me and gives useful feedback

                        Jon Wright, for writing me!
                       """

    def credits():
        showinfo("Credits", ImageD11credits)

    # GPL is stored in ImageD11/license.py as a string to be
    # displayed in the GUI if the user asks to see it


    def showlicense():
        win = ScrolledText(Toplevel(), width=100)
        win.insert(END, license)
        win.pack(expand=1, fill=BOTH)
        win.focus_set()

    # Inherits from the GuiMaker and uses functions defined above
    class TestGuiMaker(GuiMaker):
        guicommand.RETURN_NUMERICS = True
        guicommander = guicommand.guicommand()

        def start(self):
            """
            Override the GuiMaker start
            These are things to do when the gui starts
            eg: show a message about the license and list of things to do
            Then build the actual gui
            """
            startmessage = """
  ImageD11 version %s, Copyright (C) 2005-2017 Jon Wright
  ImageD11 comes with ABSOLUTELY NO WARRANTY; for details select help,
  license. This is free software, and you are welcome to redistribute it
  under certain conditions

  Please send useful feedback to wright@esrf.fr
  """ % (__version__)

            startmessage += """
  You are using version %s

  There have been lots of changes recently!

  I would also be happily surprised if it is currently working.
  """ % (__version__)
            showinfo("Welcome to ImageD11 " + __version__,
                     startmessage)

            # For the peaksearch menu

            self.peaksearcher = guipeaksearch.guipeaksearcher(self)


            self.transformer = guitransformer.guitransformer(self)

            # For the indexing - supposed to generate orientations from the
            # unitcell and g-vectors
            self.indexer = guiindexer.guiindexer(self)

            self.solver = guisolver.guisolver(self)

            # Configure the menubar (lists of Tuples of (name,
            # underline_char, command_or_submenu) )
            self.menuBar = [("File", 0,
                             [("Print", 0, self.printplot),
                              ("Exit", 1, sys.exit)]),
                            self.peaksearcher.menuitems,
                            self.transformer.menuitems,
                            self.indexer.menuitems,
                            self.solver.menuitems,
                            ("Plotting", 0,
                             [("Autoscale", 0, self.autoscaleplot),
                              ("Clear plot", 0, self.clearplot),
                              ]),
                            ("Help", 0,
                             [("Help Me!", 0, help),
                              ("History", 1, self.history),
                              ("Credits", 0, credits),
                              ("License", 0, showlicense)
                              ])]

        # The twodplot object should be taking care of it's own menu
        # Stop doing it here - TODO
        def history(self):
            win = ScrolledText(Toplevel(), width=100)
            history = self.guicommander.gethistory()
            win.insert(END, history)
            win.pack(expand=1, fill=BOTH)
            win.focus_set()

        def printplot(self):
            """
            Print the 2D plot (probably to a file?)
            """
            self.twodplotter.printplot()

        def autoscaleplot(self):
            """
            Autoscale the plot
            """
            self.twodplotter.autoscale()

        def clearplot(self):
            """
            Clear out the twodplot
            """
            self.twodplotter.clear()

        def makeWidgets(self):
            """
            Draw the gui and initialise some hidden things
            """
            # TODO Get size of TopLevels window and position it in a
            # sensible way
            #
            # Opening and saving file widgets, normally hidden, they
            # remember where you were for working directories
            self.opener = filedialog.Open()
            self.saver = filedialog.SaveAs()
            #
            # Draw the twodplot
            self.twodplotter = twodplot.twodplot(self)
            self.twodplotter.pack(side=RIGHT, expand=1, fill=BOTH)

    # Start up Tkinter
    root = Tk()
    root.wm_title("ImageD11")
    # Instantiate an object of the class TestGuiMaker
    TestGuiMaker()
    # Thats it!
    root.mainloop()




from __future__ import print_function

"""
To be moved to fabio sometime
"""

import fabio.file_series
import fabio.fabioimage
import fabio.openimage
import numpy, h5py
import gzip, bz2
from ImageD11 import ImageD11options

def get_options(parser):

    parser.add_argument("-5","--hdf5",action="store", type=str,
                      dest = "hdf5", default = None,
                      help = "hdf file containing input image series")
    # or, eventually:
    # stem, first, last, format, (omegas better be in the headers)
    parser.add_argument("-n","--stem",action="store", type=str,
                      dest = "stem", default = None,
                      help = "stem name for input image series")
    parser.add_argument("-f","--first",action="store", type=int,
                      dest = "first", default = None,
                      help = "first number for input image series")
    parser.add_argument("-l","--last",action="store", type=int,
                      dest = "last", default = None,
                      help = "last number for input image series")
    parser.add_argument("--ndigits", action="store", type=int,
                      dest = "ndigits", default = 4,
                      help = "Number of digits in file numbering [4]")
    parser.add_argument("-P", "--padding", action="store",
                      choices=["Y","N"], 
                      default="Y", dest="padding",
                      help="Is the image number to padded Y|N, e.g. "\
                          "should 1 be 0001 or just 1 in image name, default=Y")
    parser.add_argument("-F","--format",action="store", type=str,
                      dest = "format", default = ".edf",
                      help = "format [.edf] for input image series")

    parser.add_argument("-O", "--flood", action="store", 
                        type=ImageD11options.ImageFileType(mode='r'),
                      dest = "flood", default = None,
                      help = "Flood")
    
    parser.add_argument("-d", "--dark", action="store", 
                      dest = "dark", default = None,
                      type=ImageD11options.ImageFileType(mode='r'),
                      help = "Dark image")
    parser.add_argument("-S", "--step", action="store", type=float,
                      dest = "OMEGASTEP", default = None,
                      help = "omega step size")
    parser.add_argument("-T", "--start", action="store", type=float,
                      dest = "OMEGA", default = None,
                      help = "start omega")
    parser.add_argument("--omega_motor", action="store", type=str,
                      dest = "omegamotor", default = "Omega",
       help = "Header value to use for rotation motor position [Omega]")
    parser.add_argument("--omega_motor_step", action="store", type=str,
                      dest = "omegamotorstep", default = "OmegaStep",
       help = "Header value to use for rotation width [OmegaStep]")

    return parser


def get_series_from_hdf( hdf_file, dark = None, flood = None ):
    groups = hdf_file.listnames()
    for group in groups:
        imagenames = hdf_file[group].listnames()
        for image in imagenames:
            im = hdf_file[group][image]
            om = float(im.attrs['Omega'])
            data = im[:,:]
            if (dark, flood) is not (None, None):
                data = data.astype(numpy.float32)
            if dark is not None:
                numpy.subtract( data, dark, data )
            if flood is not None:
                numpy.divide( data, flood, data )
            yield fabio.fabioimage.fabioimage( data = data,
                                               header = {
                    'Omega': om } )

def series_from_fabioseries( fabioseries, dark, flood, options ):
    
    for filename in fabioseries:
        try:
            fim = fabio.openimage.openimage(filename)
        except:
            print("Missing image",filename)
            continue
        if (dark, flood) is not (None, None):
            fim.data = fim.data.astype(numpy.float32)
        if dark is not None:
            numpy.subtract( fim.data, dark, fim.data )
        if flood is not None:
            numpy.divide( fim.data, flood, fim.data )
        if options.omegamotor in fim.header:
            fim.header['Omega'] = float(fim.header[options.omegamotor])
            try:
                fim.header['OmegaStep'] = float(fim.header[options.omegamotorstep])
            except:
                fim.header['OmegaStep'] = float(options.OMEGASTEP)            
        else:
            fim.header['Omega'] = float(options.OMEGA)
            fim.header['OmegaStep'] = float(options.OMEGASTEP)            
            options.OMEGA = float(options.OMEGA) + float(options.OMEGASTEP)
        yield fim
        


def get_series_from_stemnum( options, args, dark = None, flood = None ):
    """
    Returns a file series thing - not a fabio one
    """
    if options.format in ['bruker', 'BRUKER', 'Bruker']:
        extn = ""
    elif options.format == 'GE':
        extn = ""
    else:
        extn = options.format
        
    fso = fabio.file_series.numbered_file_series(
        options.stem,
        options.first,
        options.last,
        extn,
        digits = options.ndigits,
        padding = options.padding )
    return series_from_fabioseries( fso , dark, flood, options )
    

def get_series_from_options( options, args ):
    """
    Returns a file series thing - not a fabio one

    This gives back a fabioimage object with dark and flood
    corrected data
    """

    try:
        if options.dark is not None:
            dark = fabio.openimage.openimage( options.dark ).data
        else:
            dark = None
    except:
        print("Problem with your dark",options.dark)
        raise
    
    try:
        if options.flood is not None:
            flood = fabio.openimage.openimage( options.flood ).data
        else:
            flood = None
    except:
        print("Problem with your flood",options.flood)
        raise
        

    if len(args) > 0 :
        # We assume unlabelled arguments are filenames 
        fso = fabio.file_series.file_series(args)
        return series_from_fabioseries( fso, dark, flood, options )

    if options.hdf5 is not None:
        hf = h5py.File(options.hdf5)
        # print "Getting images from",options.hdf5
        return get_series_from_hdf( hf, dark, flood )
    
    return get_series_from_stemnum( options, args,
                                     dark, flood) 
    


def getedfheader(filename):
    """
    Reads a header from an edf file in 1024 byte chunks.
    Assumes enclosing { }
    Returns string enclosed
    Adds a filename key at the top
    """
    h = "filename = "
    if filename[-3:]==".gz":
        fp=gzip.GzipFile(filename,"rb")
    elif filename [-4:]==".bz2":
        fp=bz2.BZ2File(filename,"rb")
    else:
        try:
            fp=open(filename,"rb")
        except IOError:
            return ""
    h=h+filename+";\n"
    s=fp.read(1024)
    if s.find("{")==-1:
        raise Exception("Not an edf file")
    while 1:
        if s.find("}")>=0:
            h=h+s[0:s.find("}")+2]
            break
        else:
            h=h+s
        s=fp.read(1024)
    return h

def motor_mne(hd):
    """
    expands the _mne and _pos header items of edf headers
    """
    h = {}
    order = []
    for line in hd.split(";"):
        try:
            key,vals = line.split("=")
        except ValueError:
            continue
        key = key.lstrip().rstrip()
        h[key] = vals.split(";")[0]
        order.append( key )
    for k in order:
        if k.endswith("_mne"):
            stem = k.split("_")[0]
            p = k.replace("_mne","_pos")
            newkeys = h[k].split()
            newvals = h[p].split()
            for ik, iv in zip(newkeys, newvals):
                kk = stem+":"+ik
                h[kk]=iv
                order.append( kk )
    return h, order




import unittest
import numpy as np
from ImageD11 import cImageD11

# print(cImageD11.cImageD11_avx.__file__)

def py_coo( msk ):
    i, j = np.mgrid[0:msk.shape[0], 0:msk.shape[1]]
    return i[msk].astype(np.uint16), j[msk].astype(np.uint16)


class test_clean_mask( unittest.TestCase ):
    def setUp( self ):
        shape = (9,11)
        data = np.zeros( shape, np.bool )
        data[3:5, 3:5] = 1 # keep
        self.target = data.copy()
        data[1,1] = 1 # isolated, to be removed
        data[3,0] = 1
        data[-1,-1] = 1
        data[-1,5] = 1
        data[5,-1] = 1
        data[6,0]  = 1 # adjacent to prev in mem but on edge
        self.src = data.copy()
    def test1(self):
        testmask = np.zeros( self.src.shape, np.int8 )
        npx = cImageD11.clean_mask( self.src.astype(np.int8), testmask )
        if 0:
            print('npx=',npx)
            print( self.src)
            print(self.src.view(dtype=np.int8))
            print(testmask)
            print(testmask.sum())
        self.assertTrue( self.target.sum() == npx )
        self.assertTrue( (testmask == self.target).all() )
    def test_coo(self):
        npx = self.target.sum()
        i = np.empty( npx, np.uint16)
        j = np.empty( npx, np.uint16)
        tmp = np.empty( self.target.shape[0], np.int32 )
        ok = cImageD11.mask_to_coo( self.target.astype(np.int8), i, j, tmp )
        self.assertTrue( ok == 0 )
        pi, pj = py_coo( self.target )
        self.assertTrue( ( pi == i ).all() )
        self.assertTrue( ( pj == j ).all() )
    def test_coo2(self):
        npx = self.src.sum()
        i = np.empty( npx, np.uint16)
        j = np.empty( npx, np.uint16)
        tmp = np.empty( self.target.shape[0], np.int32 )
        ok = cImageD11.mask_to_coo( self.src.astype(np.int8), i, j, tmp )
        self.assertTrue( ok == 0 )
        pi, pj = py_coo( self.src )
        self.assertTrue( ( pi == i ).all() )
        self.assertTrue( ( pj == j ).all() )
        

if __name__=="__main__":
    unittest.main()

        





from __future__ import print_function


"""
Classes for finding lattice vectors

We assume a very large unit cell
Compute the Patterson Function using this cell
Origin peaks show up for our crystals
These will be real space vectors
"""

import logging, time, sys
import numpy as np

from ImageD11 import labelimage, cImageD11, columnfile

def get_options(parser):
    parser.add_argument( '-n', '--ngrid',
                       action = 'store',
                       dest = 'npx',
                       type = int,
                       help = 'number of points in the fft grid [128]',
                       default = 128 )
    parser.add_argument( '-r', '--max_res',
                       action = 'store',
                       dest = 'mr',
                       type = float,
             help = 'Maximum resolution limit for fft (d-spacing) [1.0]',
                       default = 1.0)
    parser.add_argument( '-s', '--nsig',
                       action = 'store',
                       dest = 'nsig',
                       type = float,
             help = 'Number of sigma for patterson peaksearch threshold [5]',
                       default = 5)
    return parser


def refine_vector( v, gv, tol=0.25, ncycles=25, precision=1e-6 ):
    """
    Refine a (single) real space lattice vector
    Input
      v  :  real space lattice vector
      gv : scattering vectors (no errors)
      tol : tolerance for initial assignment to peak
      ncycles : how much to refine
      precision : when to stop refining in v += s when |s|/|v| < precision
    Returns
      vref : refined real space lattice vector

    Uses:
      hr = v[0]*gv[i,0] + v[0]*gv[i,1] + v[1]*gv[i,2]  1 per peak
      hi = round(hr) = calc                            1 per peak
      diff = calc - hr                                 1 per peak
      gradient = d(diff)/dv[i] = gv[j,i]               3 per peak
      matrix = gradient[i].gradient[j]                 3x3 peaks^2
      rhs = gradient . diff                            3

    Solve for shifts and iterates reducing tolerance as 1/(ncycles+1)
    Peaks that change index during refinement are removed
    Stops on : 
       no peaks reassigned
       ncycles runs out
       |shift| / |vec| < precision
    """
    assert gv.shape[1] == 3
    vref = v.copy()
    mg = (gv*gv).sum(axis=1)
    mold = None
    wt = 1./(mg + mg.max()*0.01)
    gvt = gv.T.copy()
    for i in range(ncycles):
        hr = np.dot( vref, gvt )   # npks
        hi = np.round( hr )         # npks
        diff = hi - hr              # npks
        m = abs(diff) < tol/(i+1)   #  keep or not ?
        diff = (diff*wt)[m]         # select peaks used
        grad = (gvt*wt).T[m]
        gvt = gvt[:,m]
        wt = wt[m]
        # lsq problem:
        rhs = np.dot( grad.T, diff ) 
        mat = np.dot( grad.T, grad )
        # avoid problems if singular:
        U,s,V = np.linalg.svd( mat )
        one_over_s = np.where( s/s.max() < 1e-6, 1, 1./s )
        mati = np.dot( U, np.dot(np.diag( one_over_s ), V ) )
        vshft = np.dot( mati, rhs )
        vref = vref + vshft
        if m.sum()==0:
            break
        slen = np.sqrt((vshft * vshft).sum())
        vlen = np.sqrt((vref*vref).sum())
        if vlen < precision or abs(slen/vlen) < precision:
            break
    return vref
    

class grid:
    def __init__(self, npx = 128, mr = 1.0 ,nsig = 5):
        """
        Set up the grid to use (the large unit cell)
        npx - number of points in the grid
        mr - maximum resolution limit to consider
        """
        self.npx = npx
        self.nsig = nsig
        self.grid = np.zeros((npx,npx,npx),np.float32)
        self.old_grid = np.zeros((npx,npx,npx),np.float32)
        self.cell_size = npx * mr / 2.
        logging.info("Using an FFT unit cell of %s"%(str(self.cell_size)))

    def gv_to_grid_new(self, gv):
        """
        Put gvectors into our grid
        gv - ImageD11.indexing gvectors
        """
        # Compute hkl indices in the fft unit cell
        logging.info("Gridding data")
        self.gv = gv
        hrkrlr = self.cell_size * gv
        hkl = np.round( hrkrlr ).astype(np.int)
        # Filter to have the peaks in asym unit
        # ... do we need to do this? What about wrapping?
        hmx = hkl.max(axis=1) <  (self.npx/2. - 2.)
        hmn = hkl.min(axis=1) >  (2. - self.npx/2.)
        my_g = np.compress( hmx & hmn, gv, axis=0 )
        # Compute hkl indices in the fft unit cell using filtered peaks
        hrkrlr = self.cell_size * my_g
        # Integer part of hkl (eg 1.0 from 1.9)
        hkl = np.floor(hrkrlr).astype(np.int)
        # Fractional part of indices ( eg 0.9 from 1.9)
        remain = hrkrlr - hkl
        grid = self.grid
        start = time.time()
        # Loop over corners with respect to floor corner
        ng = grid.shape[0]*grid.shape[1]*grid.shape[2]
        flatgrid = grid.reshape( ng )
        for cor in [ (0,0,0),   #1
                     (1,0,0),   #2
                     (0,1,0),   #3
                     (0,0,1),   #4
                     (1,1,0),   #5
                     (1,0,1),   #6
                     (0,1,1),   #7
                     (1,1,1) ]: #8
            # The corner
            thkl = hkl + cor
            fac = 1 - abs(remain - cor)
            vol = abs(fac[:,0]*fac[:,1]*fac[:,2]).astype(np.float32)
            thkl = np.where(thkl < 0, self.npx + thkl, thkl)
            ind = thkl[:,0]*grid.shape[1]*grid.shape[2] + \
                  thkl[:,1]*grid.shape[1] + \
                  thkl[:,2]

            cImageD11.put_incr( flatgrid , ind.astype(np.intp), vol )
        logging.info("Grid filling loop takes "+str(time.time()-start)+" /s")



    def fft(self):
        """ Compute the Patterson """
        start = time.time()
        self.patty = np.ascontiguousarray(abs(np.fft.fftn(self.grid)),
                                          np.float32)
        logging.info("Time for fft "+str(time.time()-start))
        self.origin = self.patty[0,0,0]
        logging.info("Patterson origin height is :"+str(self.origin))

    def props(self):
        """ Print some properties of the Patterson """
        logging.info("Patterson info "+str(self.patty.shape)
                     +str(type(self.patty)))
        p = np.ravel(self.patty)
        m = np.mean(p)
        logging.info("Average: %f"%(m))
        p2 = p*p
        self.mean = m
        v = np.sqrt( (np.sum(p2) - m*m*len(p) ) /(len(p)-1) )
        logging.info("Sigma: %f"%(v))
        self.sigma = v

    def peaksearch(self, peaksfile):
        """ Peaksearch in the Patterson """
        lio = labelimage.labelimage( (self.npx, self.npx),
                                     peaksfile )
        logging.info("Peaksearching at %f sigma"%(self.nsig))
        thresh = self.mean + self.nsig  * self.sigma
        for i in range(self.npx):
            lio.peaksearch(self.patty[i],
                           thresh,
                           i)
            lio.mergelast()
        lio.finalise()

    def pv(self, v):
        """ print vector """
        return ("%8.4f "*3)%tuple(v)

    def reduce(self, vecs):
        raise Exception("You want lattice_reduction instead")

    def read_peaks(self, peaksfile):
        """ Read in the peaks from a peaksearch """
        start = time.time()
        colf = columnfile.columnfile(peaksfile)
        logging.info("reading file %f/s"%(time.time()-start))
        # hmm - is this the right way around?
        self.rlgrid = 1.0*self.cell_size/self.npx
        self.px = colf.omega
        self.px = np.where(self.px > self.npx/2 ,
                           self.px - self.npx  ,
                           self.px)*self.rlgrid
        self.py = colf.sc
        self.py = np.where(self.py > self.npx/2 ,
                           self.py - self.npx  ,
                           self.py)*self.rlgrid
        self.pz = colf.fc
        self.pz = np.where(self.pz > self.npx/2 ,
                           self.pz - self.npx   ,
                           self.pz)*self.rlgrid
        self.UBIALL = np.array( [self.px, self.py, self.pz] ).T
        logging.info("Number of peaks found %d  %f/s, now fit some"%(
                     self.px.shape[0],time.time()-start))
        for i in range(colf.nrows):
            print(".",end="")
            self.UBIALL[i] = refine_vector( self.UBIALL[i], self.gv )
            
        logging.info("Fitting vectors %f /s"%(time.time()-start))
        self.colfile = colf
        
    def slow_score(self):
        logging.info("running slow_score")
        import time
        start = time.time()
        scores = np.dot( self.UBIALL, np.transpose( self.gv ) )
        scores_int = np.floor( scores + 0.5).astype(np.int)
        diff = scores - scores_int
        nv = len(self.UBIALL)
        print("scoring",nv,time.time()-start)
        self.tol = 0.1
        scores = np.sqrt(np.average(diff*diff, axis = 1))
        n_ind = np.where(np.absolute(diff)< self.tol, 1 , 0)
        nind = np.sum(n_ind, axis=1)
        order = np.argsort(nind)[::-1]
        mag_v = np.sqrt( self.px * self.px +
                        self.py * self.py +
                        self.pz * self.pz )
        f = open("fft.pks","w")
        for i in range(nv):
            j = order[i]
            f.write("%d %9.5f %9.5f %9.5f %9.5f %9.5f %9.5f %7d "%(
                i,
                self.px[j],
                self.py[j],
                self.pz[j],
                mag_v[j],
                scores[j],
                self.colfile.sum_intensity[j],
                nind[j]
                ))
            for k in range(i+1,nv):
                l = order[k]
                nij = np.sum( n_ind[j] * n_ind[l] )
                f.write("%4d : %-7d "%(k,nij))
            f.write("\n")

        f.close()
        print(diff.shape)
        return diff
###################################################
## sum_sq_x = 0
## sum_sq_y = 0
## sum_coproduct = 0
## mean_x = x[1]
## mean_y = y[1]
## for i in 2 to N:
##     sweep = (i - 1.0) / i
##     delta_x = x[i] - mean_x
##     delta_y = y[i] - mean_y
##     sum_sq_x += delta_x * delta_x * sweep
##     sum_sq_y += delta_y * delta_y * sweep
##     sum_coproduct += delta_x * delta_y * sweep
##     mean_x += delta_x / i
##     mean_y += delta_y / i
## pop_sd_x = sqrt( sum_sq_x / N )
## pop_sd_y = sqrt( sum_sq_y / N )
## cov_x_y = sum_coproduct / N
## correlation = cov_x_y / (pop_sd_x * pop_sd_y)
###################################################






        # sm = np.zeros( (len(diff), len(diff)), np.float)
        # for k in range(len(diff)):
        #     i = order[k]
        #     sm[i,i] = np.dot(diff[i], diff[i])
        # for k in range(len(diff)-1):
        #     i = order[k]
        #     for l in range(i+1, len(diff)):
        #         j = order[l]
        #         sm[i,j] = np.dot(diff[i],diff[j])/sm[i,i]/sm[j,j]
        #         sm[j,i] = sm[i,j]
        # for i in range(len(diff)):
        #     sm[i,i] = 1.
        # print(sm[:5,:5])
        # print("Scoring takes",time.time()-start)
        # return sm

#####
#        To Do
#        Convert the labelimage/columnfile to write to memory
#        process / score / group the peaks after peaksearching
#
#        --- Separate program needed now?
#            foreach peak - is it a new peak, or sum of previous ones
#                         - how many gvecs does it make integers of
#                         - which gvecs does it make integers of w.r.t others
#             ngvecs * npeaks == 175 * 200,000 K - 10^7 = 10Mpixel image ?
#

def test(options):
    gvfile = options.gvfile
    mr = options.max_res
    npx = options.npx
    nsig = options.nsig
    from ImageD11 import indexing
    print(npx, mr, nsig)
    go = grid(npx = npx , mr = mr , nsig = nsig)
    io = indexing.indexer()
    io.readgvfile(gvfile)
    go.gv_to_grid_new(io.gv )
    go.fft()
    go.props()
    go.peaksearch(open(gvfile+".patterson_pks","w"))
    im = go.read_peaks(gvfile+".patterson_pks")
    #print "Before reduction", len(go.UBIALL)
    #sys.stdout.flush()
    #ubinew = go.reduce(go.UBIALL)
    #print "After reduction", len(ubinew)
    go.UBIALL = ubinew
    go.slow_score()
    #from matplotlib.pylab import imshow, show
    #imshow(im)
    #show()
    #return im, go
    sys.exit()


    go.gv_to_grid_old(io.gv)

    diff = np.ravel(go.grid - go.old_grid)
    print("All OK if this is zero",np.add.reduce(diff))
    print("error at",np.argmax(diff),diff.max(),np.argmin(diff),diff.min())





if __name__=="__main__":

    import sys
    class options:
        max_res = 0.7
        npx = 128
        nsig = 10

    options.gvfile = sys.argv[1]

    test(options)




from __future__ import print_function, division
from ImageD11 import columnfile, parameters, transform, cImageD11

import numpy as np
import sys, time

class binned_array(object):
    def __init__(self, ar, nbins=255 ):
        """
        Digitise an array onto nbins precision
        """
        self.ar = np.asarray( ar )
        self.nbins = nbins
        lo = self.ar.min()
        hi = self.ar.max()
        self.step = (hi - lo) / (nbins - 1)
        self.low = lo - self.step/2
        self.ibin = self.bin( self.ar )
        self.bc = np.bincount( self.ibin )
        cs = np.cumsum( self.bc )
        self.bin_inds = np.concatenate( ([0,], cs) )
        
    def bin(self, x):
        """Assign x values into bins """
        return np.round( (x - self.low) / self.step ).astype( int )
        

def absanglediff( a1 , a2 ):
    ra1 = np.radians( a1 )
    ra2 = np.radians( a2 )
    er = np.arctan2( np.sin( ra1 - ra2 ), np.cos( ra1 - ra2 ) )
    return abs(np.degrees( er ))


def findpairs_2d( cf, dangle, omestep, fac = 1.5):
    """
    Search for Friedel pairs (and harmonics) as parallel (or nearly)
    vectors

    Steps:
      1) Compute g-vectors 
      2) Normalise them to get directions
      3) locate them onto the 2D surface of directions (sphere in 3d)
      4) "flip sign" to get pairs of (g, -g) to overlap
      5) foreach 2D neighborhood look for |g|, 2|g|, 3|g|, etc.

    cf contains gx, gy, gz already
    dangle is the twotheta and eta precision
    omestep is the omega step
    """
    cf.sortby( "ds" ) # sort on mod_g for later peak groupings
    #
    # Normalise the g-vectors to give direction cosines
    nve =  np.array((cf.gx, cf.gy, cf.gz)) / cf.ds 
    # cut on y > 0 (could equally be x... )
    nve = np.where( nve[1] > 0, nve, -nve )
    # World map geometry ...
    #     angle to z-axis   n.z = |n||z|cos( latitude )
    az = np.degrees( np.arccos( nve[2] ) )
    #     angle to xy-axis  x/y = tan( longitude )
    ay = np.degrees( np.arctan2( nve[0], nve[1] ) )
    # Make bins reflect the estimated precision (not ideal for now)
    # ... guess latitude error is also eta error, make bins accordingly
    nz = int( 180.0/(dangle*fac) )
    zstep = 180.0/nz
    zedges = np.arange( az.min() - zstep/2, az.max() + zstep, zstep )
    # ... and longitude errors is roughly omega error
    ny = int( 180/(omestep*fac) )
    ystep = 180.0/ny
    yedges = np.arange( ay.min() - ystep/2, ay.max() + ystep, ystep )
    #
    iz = (az - zedges[0])*( len(zedges) - 1 ) / ( zedges[-1] - zedges[0]  )
    iy = (ay - yedges[0])*( len(yedges) - 1 ) / ( yedges[-1] - yedges[0]  )
    iz = np.floor(iz).astype( int )
    iy = np.floor(iy).astype( int )

    # Collect all the spots into a histogram in 2D
    start = time.time()
    H, zHedges, yHedges = np.histogram2d( az, ay, ( zedges, yedges ) )
    H = H.astype( np.int )
    NZ, NY = H.shape
    if __debug__: # checkng how binning worked
        print("H2D:",time.time()-start)
        assert (zHedges == zedges).all()
        assert (yHedges == yedges).all()
        #     v   - minbin
        # Find which bin the sqpots are in
        assert (iz == np.digitize( az, zedges )-1).all(),(
            iz,np.digitize( az, zedges ))
        assert (iy == np.digitize( ay, yedges )-1).all()
        # H is ordered as    H.shape[0]   H.shape[1]
        #print(H.shape   ,zedges.shape, yedges.shape )
        #print( iz.min(), iz.max(), iy.min(), iy.max() )
        #b = np.argmax(iz)
        #print(az[b],iz[b],zedges[-1])
        #b = np.argmax(iy)
        #print(ay[b],iy[b],yedges[-1])
    #
    # Find the storage locations to sort the peaks onto the 2D histogram
    pointers = np.cumsum( H.ravel(), dtype = np.int64 )
    # Not sure we have a compiled code for this ?
    # loop over peaks, make location be given by pointer
    #  ... increment pointer to make space for next
    # first bin contains 2 peaks. Place these at 0,1. Next starts at 2.
    ptemp = np.zeros(pointers.shape, np.int )
    ptemp[0] = 0
    ptemp[1:] = pointers[:-1]
    # To index these pointers
    pidx = iz*NY + iy
    # resulting sort
    order = np.zeros( cf.nrows, int )
    start = time.time()
    for i in range(cf.nrows):
        p = pidx[i]           # p = which pixel is this peak
        order[i] = ptemp[p]   # output position
        ptemp[p] += 1
    # sort the columnfile and histogram indices
    indices = np.arange( cf.nrows, dtype=np.int )
    indices.put( order, indices.copy())
    cf.reorder( indices )
    iy = iy[indices]
    iz = iz[indices]
    pidx = iz*NY + iy
    if __debug__:
        print('splatting', time.time()-start )        
        assert len(np.unique(order)) == cf.nrows
        assert order.min() == 0
        assert order.max() == cf.nrows-1
    start = time.time()
    allpointers = np.concatenate( ( [0,], pointers ) )
    pimage = np.reshape( pointers, H.shape )
    # Now the peaks are sorted according to which direction they have
    # each peak has an i,j co-ordinate showing where it is on the world map
    # we want to select peaks for each pixel
    izimage, iyimage = np.mgrid[ 0:H.shape[0], 0:H.shape[1] ]
    mask = H > 0
    k=0
    hits = []
    for iiy, iiz in zip( iyimage[mask], izimage[mask] ) :
        # print("Pixel",iiy, iiz)
        # In the y direction we will wrap around
        newhits = [] 
        if iiz <= 1 or iiz > H.shape[0]-2: # skip edges
            continue
        for this_y in (iiy-1)%NY , iiy, (iiy+1)%NY:
            # On the z direction we do not wrap as this is a pole
            #  (fable geometry)
            for this_z in  iiz - 1, iiz, iiz + 1: 
                pkhere = H[this_z, this_y]
                if pkhere == 0:
                    continue
                # location of these peaks is:
                plo = allpointers[ this_z * NY + this_y ]
                phi = allpointers[ this_z * NY + this_y + 1]
                newhits += range(plo, phi )
        if len(newhits)>1:
            hits.append(newhits)
    print(len(hits), "hits, %.5f s"%(time.time()-start))
    if __debug__:
        # assert order is a permutation ...
        assert len( np.unique(order)  ) == cf.nrows, (
            len( np.unique(order)  ), cf.nrows)
        assert order.min() == 0
        assert order.max() == cf.nrows-1
    if 0:
        import pylab as pl
        pl.figure(1)
        pl.imshow( H, interpolation='nearest',
                   origin='low',
                   extent=( yedges[0], yedges[-1], zedges[0], zedges[-1] ),
                   aspect = 'auto',
                   )
        pl.plot( ay, az, "+")
        pl.show()
    start = time.time()
    #
    # convert dangle error to error in ds
    # ... Bragg :    w  = 2*d*sin(theta)     w = wvln, d=d-spacing
    #                ds = 2*sin(theta)/w     theta = Bragg angle
    #                                        ds = 1/d = dstar
    #     d(ds)/d(theta) = 2*cos(theta)/w    since d(sin) = d
    #     d(ds) = 2*cos(theta)*d(theta)/w
    #     d(ds) = 2*cos(theta)*(d(2theta)/2)/w
    #     d(ds) = cos(theta)*d(2theta)/w
    # take cos(theta) == 1
    w = cf.parameters.get('wavelength')
    d_ds = np.radians(dangle) / w
    print("Converting dangle",dangle, "degrees to",d_ds, "ds_tol")
    for p,h in enumerate(hits):
        clusterpeaks( cf, h, d_ds )
        if p % 100 == 0:
            print("%8d %.2f"%( p, 100*p/len(hits)) , end="\r" )
    if __debug__:
        print( "Time clustering",time.time()-start )
            

def cluster1d( ar, tol ):
    """
    try to figure out what goes together
    ar  = 1d array
    tol = order of magnitude to say things are the same

    returns array of cluster indices:
       [ 0,0, 2,2, 5,5, 1, 2,2, 0,0] etc
    """
    idx = [[0,]]
    for i,v in enumerate(ar[1:]):
        new_item = True
        for j in range(len(idx)):
            if abs(v - ar[idx[j][0]]) < tol:
                idx[j].append( i+1 )
                new_item = False
                break
        if new_item:
            idx.append([ i+1,] )
    return idx
    
def clusterpeaks( cf, pks, d_ds ):
    """
    Cases : 
        1)  Identical peaks (tth, eta, omega) with different dty
            ... sort by tth and then omega
        x)  Friedel pairs (gve, -gve)
            ... g1 + g2 == 0
            ... is a special case of:
        2)  Harmonics
            ... g2 = 2*g1
            ... |g2| / |g1| == integer and direction(g1)==direction(g2)
    """
    assert len(pks) > 1
    # better to sort on ds and start with the smallest first ?
    ds    = cf.ds[  pks]
    order = np.argsort( ds ).astype( np.intc )
    ids = np.zeros( len(pks), np.intc )
    avgs = np.zeros( len(pks), np.float )
    nclusters = cImageD11.cluster1d( ds , order, d_ds, ids, avgs)
    # print(ds)
    # print( nclusters, ids, avgs )                        
    if 0:
        print("# npks",len(pks),end=" ")
        print("nclusters:",nclusters)
        print("Average d spacings",avgs[:nclusters])
    
    # now identify "equivalence" or "gaps" in omega
    #  ... then identify if they are harmonics

    
    # START HERE : what is the output from this process....
    #    averaged_peak (sc, fc, etc)  ... identical peaks (dty),
    #                                       list of contributors
    #    harmonics...     m=1, m=2 etc

def unit(vec):
    return vec/np.sqrt(np.dot(vec,vec))


def skewlines( o1, v1, o2, v2 ):
    # distance between two skew lines
    n = np.cross(v1, v2)
    mag = np.sqrt( np.dot( n, n ) )
    dv = o1 - o2
    if mag > 0:
        un = n / mag
        dv = np.dot( un, dv )
    else:
        # lines are parallel, distance is constant
        pass
    return np.sqrt( np.dot( dv, dv ) )
            

def fitpair( cf, pair, pars, sol0=None ):
    # pair is a pair of peaks with g, -g
    i, j = pair
    if sol0 is None:
        tx0 = 0.
        ty0 = 0.
        tz0 = 0.
        t0 = np.zeros(3)
    else:
        tx0 , ty0, tz0 = sol0
        t0 = sol0
    xyz = np.array(( ( cf.xl[i], cf.yl[i], cf.zl[i]),
                     ( cf.xl[j], cf.yl[j], cf.zl[j]) ) ).T
    om = np.array((cf.omega[i], cf.omega[j]))
    tth0, eta0 = transform.compute_tth_eta_from_xyz(
        xyz, om, t_x=tx0, t_y=ty0, t_z=tz0,
        wedge=pars.get('wedge'), chi=pars.get('chi'))
    g0 = transform.compute_g_vectors(
        tth0, eta0, om, pars.get("wavelength"),
        wedge=pars.get('wedge'), chi=pars.get('chi'))
    tth1, eta1 = transform.compute_tth_eta_from_xyz(
        xyz, om, t_x=tx0+1, t_y=ty0, t_z=tz0,
        wedge=pars.get('wedge'), chi=pars.get('chi'))
    g1 = transform.compute_g_vectors(
        tth1, eta1, om, pars.get("wavelength"),
        wedge=pars.get('wedge'), chi=pars.get('chi'))
    dg_dtx = g1 - g0
    tth1, eta1 = transform.compute_tth_eta_from_xyz(
        xyz, om, t_x=tx0, t_y=ty0+1, t_z=tz0,
        wedge=pars.get('wedge'), chi=pars.get('chi'))
    g1 = transform.compute_g_vectors(
        tth1, eta1, om, pars.get("wavelength"),
        wedge=pars.get('wedge'), chi=pars.get('chi'))
    dg_dty = g1 - g0
    tth1, eta1 = transform.compute_tth_eta_from_xyz(
        xyz, om, t_x=tx0, t_y=ty0, t_z=tz0+1,
        wedge=pars.get('wedge'), chi=pars.get('chi'))
    g1 = transform.compute_g_vectors(
        tth1, eta1, om, pars.get("wavelength"),
        wedge=pars.get('wedge'), chi=pars.get('chi'))
    dg_dtz = g1 - g0
    #      kx = kx          ,  ky = -ky        , kz = -kz
    gerr = g0[:,0] + g0[:,1]
    de_dtx = dg_dtx[:,0] + dg_dtx[:,1]
    de_dty = dg_dty[:,0] + dg_dty[:,1]
    de_dtz = dg_dtz[:,0] + dg_dtz[:,1]
    grad = ( de_dtx, de_dty, de_dtz )
    rhs = np.zeros(3)
    mat = np.zeros((3,3))
    if 0 and sol0 is not None:
        print(tth0)
        print(eta0)
        print(g0)
        print(gerr)
    for i in range(3):
        rhs[i] = np.dot( grad[i], gerr )
        for j in range(i,3):
            mat[i,j] = mat[j,i] = np.dot( grad[i], grad[j] )
    w , v = np.linalg.eigh( mat )
    ### mat   = v.T .  w .  v.T-1
    ### mat   = v.T .  w .  v      since v is like a U matrix (orthonormal)
    ### mat-1 = v.T . 1/w . v
    ww = np.eye(3)
    ans = np.zeros(3)
    sol = t0
    # Take the two largest eigenvalues, eigh says sorted
    assert w[0] < w[1] < w[2], "check eigh returns sorted?"
    for i in range(2,0,-1):
        ww[i,i] = 1/w[i]
        ans += np.dot(v, ww[i]*np.dot(v.T, rhs) )
        sol -= ans/w[i]
        print(i,ans,sol,w[i])
    ans = -unit(np.dot(v, ww[2]*np.dot(v.T, rhs) ))
    sol -= ans*w[2]
    print(ans,sol,w)
    imat = np.dot(np.dot(v, ww),v.T)
    #print('ima',np.dot(np.linalg.inv( mat ) , rhs ))
    return sol, ans, (w[0]/w[1],w[1]/w[2])


def guess_omega_step( colf ):
    if "Max_o" in colf.titles and "Min_o" in colf.titles:
        mask = colf.Max_o == colf.Min_o
        if mask.sum() == 0:
            return 0
        uniqomega =  np.array(list(set(colf.omega[mask])))
    else:
        uniqomega =  np.array(list(set(colf.omega)))
    uniqomega.sort()
    domegauniq = uniqomega[1:] - uniqomega[:-1]
    # mode would be better statistic ?
    med = np.median( domegauniq )
    return med
    
    
    
        

def main():
    colf = columnfile.columnfile( sys.argv[1] )
    print("Read",colf.nrows,"peaks")
    p = parameters.read_par_file( sys.argv[2] )
    # input a sample radius to use for computing ranges
    radius = int( sys.argv[3] )
    omegastep = guess_omega_step( colf )
    print("Guess omega step as", omegastep)
    colf.updateGeometry( p )

    
    for i,t in enumerate(colf.titles):
        print("\n", t, end=" ")
        for j in (0,1,2, colf.nrows//2,colf.nrows//2+1,
                  colf.nrows//2+2, colf.nrows-3, colf.nrows-2, colf.nrows-1):
            print( colf.bigarray[i,j], end=" ")
    print()

    modXL = np.sqrt(colf.xl*colf.xl + colf.yl*colf.yl + colf.zl*colf.zl)
    modXLmean = modXL.mean()
    dangle = np.degrees( np.arctan2( radius , modXLmean ) )
    print ("Angle precision is about ",dangle,"for average distance",modXLmean,
           "and sample size ",radius)

    findpairs_2d( colf, dangle, omegastep )


if __name__=="__main__":
        main()


        
# dty merging
# cd /data/id11/jon
# python ~/ImageD11/sandbox/friedel.py ma3918/25um_z000.hdf ma3918/tin_fitted.par  300
# 180 deg, larger (331611 peaks):
# python  ~/ImageD11/sandbox/friedel.py 1702/fe3o4_diff_20170212_160236_/iflip_t500.flt 1702/fe3o4_diff_20170212_160236_/fe3o4.par 1000















"""
We started using some intrinsics in the C code to make it run faster
when processors support specific instructions.  This makes it less
portable and harder to build. These notes explain how things could 
work. 

If a CPU has some feature (sse2, avx2, fma, etc) then we will want
to use it in specific places that we are aware of. We might as well
also let the compiler use those instructions everywhere else where
we have not written specific things too. That means:

- compile everything as avx2 -> avx2 version
- compile everything as sse2 -> sse2 version
- compile everything as ansi -> ansi version

So we are going to build a series of extensions and use #ifdef inside
them.

Which one we get is determined by the cImageD11.py module.
"""

import os, sys, platform
import setuptools
import distutils.ccompiler

need_build = True

sources = ("blobs.c cdiffraction.c closest.c connectedpixels.c"+
 " darkflat.c localmaxlabel.c sparse_image.c splat.c").split()

plat = platform.system()
bits = platform.architecture()[0]
vers = "%d.%d"%(sys.version_info[:2])
tmpdir = "%s_%s_%s"%(plat, bits, vers)
avx2libname = "cImageD11_"+tmpdir+"_avx"
sse2libname = "cImageD11_"+tmpdir+"_sse2"
# ansi ?

compiler = None
for a in sys.argv:
    if "mingw32" in a:
        compiler = "mingw32"



if plat == "Linux" or compiler == "mingw32":
    arg=["-O2", "-fopenmp", "-fPIC", "-std=c99" ]
    sse2arg = arg + ["-msse4.2"]
    avx2arg = arg + ["-mavx"]
    # link args
    lsse2arg = arg + ["-msse4.2"]
    lavx2arg = arg + ["-mavx"]
elif plat == "Windows":
    arg=["/Ox", "/openmp" ]
    # the /arch switches are ignored by the older MSVC compilers
    sse2arg = arg + ["/arch:SSE2",]
    avx2arg = arg + ["/arch:AVX",]
    lsse2arg = []
    lavx2arg = []
else:
    avx2arg = sse2arg = arg = [ ]

def run_cc( cc, plat, bits, vers, name, flags, libname ):
    objs = cc.compile( sources , 
                       output_dir=libname.replace("cImageD11_",""),
                       extra_preargs = flags )
    ok = cc.create_static_lib( objs, libname, output_dir="." )
    return libname

def write_docs( inp, outf ):
    """ One single block of !DOC per item, first word is key
    """
    with open(inp , "r") as pyf:
        fname = None
        docs = {}
        for line in pyf.readlines():
            if line.startswith("!DOC"):
                if fname is not None:
                    docs[fname] += line[5:] 
                else:
                    words = line.split()
                    fname = words[1]
                    docs[fname] = " ".join(words[2:]) + "\n"
            else:
                fname = None
    with open(outf, "w") as docf:
        keys = list(docs.keys())
        keys.sort()
        for fname in keys:
            docf.write( '%s = """%s"""\n'%(fname, docs[fname]))
        docf.write("__all__ = [" + ",".join(['"%s"'%(k) for k in keys ])+  "]")
        

def make_pyf( inp, name ):
    out = open(name+".pyf", "w")
    out.write("python module %s\n"%(name))
    pyf = open(inp , "r").read() 
    out.write( pyf )
    out.write("end python module %s\n"%(name))
    out.close()

def docs():
    write_docs( "cImageD11_interface.pyf", os.path.join("..","ImageD11src", "cImageD11_docstrings.py"))


def main():
    cc = distutils.ccompiler.new_compiler( verbose=1 , compiler=compiler )
    cc.add_include_dir( "." )
    docs()
    make_pyf( "cImageD11_interface.pyf", "cImageD11_sse2")
    make_pyf( "cImageD11_interface.pyf", "cImageD11_avx")
    sse2lib = run_cc(cc, plat, bits, vers, "sse2", sse2arg, sse2libname )
    avx2lib = run_cc(cc, plat, bits, vers, "avx", avx2arg, avx2libname )
    return 0


if __name__=="__main__":
    main()




import sys
from ImageD11.grain import write_grain_file, grain, read_grain_file

ubi = sys.argv[1]

grains = read_grain_file( ubi )

res = sys.argv[2]

for line in open(res).readlines():
    if line.find("pos_grains")==0:
        items=line.split()
        j = int(items[0].split("_")[-1])
        grains[j].translation = [float(x)*1000 for x in items[1:]]

write_grain_file(sys.argv[3], grains)
    




from __future__ import print_function

import os, time, glob, sys

firstrun = True

log=open("log","w")

while 1:
    edfs = glob.glob("*.edf")
    chis = glob.glob("*.chi")
    if firstrun:
        chis = []
        firstrun = False
    todo = []
    for e in edfs:
        if e.replace("edf","chi") not in chis:
            num = int(e.split("tion")[1].split(".")[0])
            todo.append( num )
    if len(edfs) != len(chis):
        os.system("fit2dcake.py -g TbAsO4_with_transition -p newfar.par")
        os.system("rm tmp*")
        first = min(todo)
        last = max(todo)
        log.write("Fitting %d %d\n"%(first,last))
        os.system("rm fits")
        os.system("rm results")
        os.system("fitem_jon_chi TbAsO4_with_transition %d %d 5.9 6.27"%(
            first, last))
        os.system("cat results >> pkfits")
    print()
    print("Control-c to exit me\r", end=' ')
    sys.stdout.flush()
    time.sleep(30)



#!/usr/bin/env python

from __future__ import print_function


from ImageD11 import opendata
import glob

def getnumerics(d):
    ret = {}
    for k,v in list(d.items()):
        try:
            ret[k] = float(v)
        except ValueError:
            pass
    return ret

import sys
        
class headerfollower:
    
    # Header items we choose to ignore changes in for printing
    
    ignore = ['time_of_day','stream700','pico1','Ring Current','samome',
              'run']
    interesting = [ 'time' ,  'dir', 'prefix', 'suffix']

    
    def __init__(self, filename):
        """
        Initialise on a filename
        """
        self.header = opendata.edfheader(filename)
        self.numerics = getnumerics(self.header)
        self.headeritems = set(self.header)
        self.filename = filename
        self.report = {}
        self.reportnames = []
        for k in self.interesting:
            self.addreport(filename, k, self.header[k])
        for h,v in list(self.numerics.items()):
            self.addreport(filename, h, v)
        
    def addreport(self, name, item, val):
        """
        Store up the things to report in a dictionary
        Holds a list of names, then item val pairs to report
        """
        if name in self.report:
            self.report[name].append( (item, val) )
        else:
            self.report[name] = [(item, val)]
            self.reportnames.append(name)

    def printreport(self):
        """
        pretty print the report information
        """
        print()
        for name in self.reportnames:
            print(name)
            sr = self.report[name]
            sr.sort()
            for k, v in sr:
                print("\t",k,v)
            print()
        print("Last was",self.filename)
                                       
    def nextfile(self,filename):
        """
        process the next file
        """
        h = opendata.edfheader(filename)
        hks = set(h.keys())
        changes = hks.symmetric_difference(self.headeritems)
        if len(changes) != 0:
            raise Exception("Filename %s introduces different header keys"%
                            (filename))
        n = getnumerics(h)
        for k in list(n.keys()):
            if k == "Omega":
                if n[k] < self.numerics[k]:
                    self.addreport(self.filename,k,self.header[k])
                    self.addreport(filename, k, h[k])
                self.numerics[k] = n[k]
                self.header[k] = h[k]
                continue
            if k in self.ignore:
                continue
            if n[k] != self.numerics[k]: # number changed
                # report
                self.addreport(self.filename,k,self.header[k])
                self.addreport(filename, k, h[k])
                self.numerics[k] = n[k]
                self.header[k] = h[k]
        self.filename = filename
        
            

print("listing edf files")
filelist = glob.glob("*.edf")

filelist = [ (opendata.getnum(name), name) for name in filelist ]

filelist.sort()

print("processing, please wait")
obj = headerfollower(filelist[0][1])

for n,f in filelist[1:]:
    obj.nextfile(f)

obj.printreport()







#!/usr/bin/env python

from __future__ import print_function

"""
Utility script to pick out the peaks belonging to a certain
grain

Used to stabilise refinements
"""

import sys, logging

import numpy as np
from argparse import ArgumentParser

from ImageD11 import refinegrains, indexing, grain, ImageD11options



def filtergrain(options):
    """
    Filter a peaks file according to which peaks are indexed
    """
    o = refinegrains.refinegrains(tolerance=0.9,
                                  OmFloat=options.omega_float,
                                  OmSlop=options.omega_slop)
    o.loadparameters(options.parfile)
    o.readubis(options.ubifile)

    if options.grain is None:
        if len(o.grainnames) == 1:
            gn = o.grainnames[0]
        else:
            for i,gn in zip(list(range(len(o.grainnames))),o.grainnames):
                logging.info("Choose %d for grain %s"%(i, gn))
            gn = o.grainnames[int(input("select which grain "))]
    else:
        gn = o.grainnames[int(options.grain)]
    o.grainnames = [ gn, ]

    o.loadfiltered(options.fltfile)
    o.generate_grains()
    o.assignlabels()
    o.set_translation( gn, options.fltfile )
    o.compute_gv(o.grains[ (gn, options.fltfile) ] )
    if options.tol is None:
        for tol in [0.01,0.025,0.05,0.1,0.15,0.25,0.5]:
            o.tolerance = tol
            logging.info("tol %f"%(o.tolerance))
            o.refineubis(quiet=False, scoreonly=True)
        o.tolerance = float(input("Enter tolerance "))
        options.tol = o.tolerance
    else:
        o.tolerance = options.tol
    matrix = o.grains[(gn,options.fltfile)].ubi

    o.assignlabels()
    drlv2 = indexing.calc_drlv2( matrix, o.gv )
    logging.info("Total peaks before filtering %d"%
                     o.scandata[options.fltfile].nrows)
    gotpks = o.scandata[options.fltfile].copy()
    gotpks.filter(gotpks.labels == gn)
    gotpks.writefile(options.newfltfile)
    logging.info("Peaks which were indexed %d written to %s"%(
                gotpks.nrows, options.newfltfile))
    # don't bother to copy here as we can overwrite
    if options.notindexed is not None:
        notpks = o.scandata[options.fltfile].copy()
        notpks.addcolumn(o.tth, "tth")
        notpks.addcolumn(o.eta, "eta")
        notpks.addcolumn(o.gv[:,0] , "gx")
        notpks.addcolumn(o.gv[:,1] , "gy")
        notpks.addcolumn(o.gv[:,2] , "gz")
        notpks.filter(drlv2 > o.tolerance*o.tolerance)
        notpks.writefile(options.notindexed)
        logging.info("Peaks which were not indexed %d written to %s"%(
                notpks.nrows, options.notindexed))
    if options.newubifile is not None:
        o.scandata[options.fltfile] = gotpks
#        matrix = o.refine(matrix,quiet=True)
        grain.write_grain_file(options.newubifile,[o.grains[(gn,options.fltfile)]])
        logging.info("Refined ubi in %s "%(
                         options.newubifile))

def get_options(parser):
    parser=refinegrains.get_options(parser)
    parser.add_argument("-N",  "--notindexed", action="store",
                      dest="notindexed", 
                      type=ImageD11options.ColumnFileType(mode='w'),
                      help="Name of flt file for unindexed peaks")
    parser.add_argument("-g","--grain", action="store",
                      dest = "grain", type=int, default=None,
                      help = "Which grain to choose")

    parser.description = """
Filtergrain should choose the peaks from a filtered
peaks output file according to those which are closest
to a particular grain
    """
    return parser

if __name__=="__main__":

    console = logging.StreamHandler(sys.stdout)
    # set a format which is simpler for console use
    formatter = logging.Formatter('%(levelname)-8s : %(message)s')
    # tell the handler to use this format
    console.setFormatter(formatter)
    # add the handler to the root logger
    console.setLevel(logging.DEBUG)
    root = logging.getLogger('')
    root.addHandler(console)
    root.setLevel(logging.DEBUG) # should we process everything...?


    parser = get_options( ArgumentParser() )

    options = parser.parse_args()

    if None in [options.parfile,
                options.ubifile,
                options.newfltfile,
                options.fltfile]:
        parser.print_help()
        logging.error("You have not filled in all the required options")
        import sys
        sys.exit()

    try:
        filtergrain(options)
    except:
        parser.print_help()
        import traceback
        logging.error("An error occurred, details follow")
        traceback.print_exc()








import glob, shutil 


def got_future(lines):
    for line in lines:
        if line.find("from __future__ import print_function")==0:
            return True
    return False

def got_print(lines):
    for line in lines:
        if line.find("print") > -1:
            return True
    return False

def got_input(lines):
    for line in lines:
        if line.find("input") > -1:
            return True
    return False

fl = glob.glob("*.py")

def fix(fname, lines):
    shutil.move(fname, fname+".bak2")
    g = open(fname,"w")
    g.write("\nfrom __future__ import print_function\n")
    if got_input(lines):
        g.write("from six.moves import input\n")
    for line in lines:
        g.write(line)
    g.close()


for fname in fl:
    lines = open(fname).readlines()
    p = got_print(lines) 
    f = got_future(lines)
    if p and f:
        print "OK",fname
        continue
    if p:
        print "Need to fix",fname
        fix(fname, lines)







from __future__ import print_function, division

import ctypes, numpy, sys, time, os
if 0: 
   os.system("gcc -shared -fPIC -O3 -mavx2 imsum.c -o imsum.so")

timer = time.perf_counter

data = (numpy.random.random( 2048*2048*32 )*10000).astype(numpy.uint16)
if "win" in sys.platform:
    lib = ctypes.CDLL( "imsum.dll" )
    cfuns = [lib.imsum_sse2_0 ,lib.imsum_sse2_1 , lib.imsum_c_simd,
            lib.imsum_c,lib.imsum_avx2_0,lib.imsum_avx2_1,lib.imsum_avx2_2]
else:
    lib = ctypes.CDLL( "imsum.so" )
    cfuns = [lib.imsum_sse2_0 ,lib.imsum_sse2_1 ]
    if "avx2" in open("/proc/cpuinfo").read():
        print("You have avx2")
        cfuns += [lib.imsum_c,lib.imsum_avx2_0,lib.imsum_avx2_1]

lib.imsum_sse2_0.argtypes = [ ctypes.POINTER( ctypes.c_uint16 ), ctypes.c_int64 ]
lib.imsum_sse2_0.restype  = ctypes.c_int64
lib.imsum_sse2_1.argtypes = [ ctypes.POINTER( ctypes.c_uint16 ), ctypes.c_int64 ]
lib.imsum_sse2_1.restype  = ctypes.c_int64
lib.imsum_avx2_0.argtypes = [ ctypes.POINTER( ctypes.c_uint16 ), ctypes.c_int64 ]
lib.imsum_avx2_0.restype  = ctypes.c_int64
lib.imsum_avx2_1.argtypes = [ ctypes.POINTER( ctypes.c_uint16 ), ctypes.c_int64 ]
lib.imsum_avx2_1.restype  = ctypes.c_int64
lib.imsum_avx2_2.argtypes = [ ctypes.POINTER( ctypes.c_uint16 ), ctypes.c_int64 ]
lib.imsum_avx2_2.restype  = ctypes.c_int64
lib.imsum_c.argtypes = [ ctypes.POINTER( ctypes.c_uint16 ), ctypes.c_int64 ]
lib.imsum_c.restype  = ctypes.c_int64
lib.imsum_c_simd.argtypes = [ ctypes.POINTER( ctypes.c_uint16 ), ctypes.c_int64 ]
lib.imsum_c_simd.restype  = ctypes.c_int64




    
val = data.sum(dtype=numpy.int64)
    
N=100
t0=timer()
vals = [ data.sum(dtype=numpy.int64) for i in range(N)]
t1=timer()
avg = (t1-t0)/N
print("numpy",avg)
for cfunc in cfuns[::-1]:
    t0=timer()
    valc = [cfunc( data.ctypes.data_as( ctypes.POINTER( ctypes.c_uint16 ) ),
                   data.shape[0] )
            for i in range(N)]
    t1=timer()
    avc = (t1-t0)/N
    ok = (numpy.array(valc) == vals).all()
    print(cfunc.__name__,"%.6f"%(avc),"%.6f"%(avg/avc),ok)
    if not ok:
        print(valc[0],vals[0])

print(len(data)*5000)

"""
hpc3-2502:~/git/ImageD11_master/src % python test_imsum.py 
You have avx2
numpy 0.00356456112862
imsum_c 0.000729 4.889183 True
imsum_sse2_0 0.001130 3.155040 True
imsum_sse2_1 0.000969 3.678305 True
imsum_avx2_0 0.000523 6.813989 True
imsum_avx2_1 0.000520 6.858841 True
"""



#!/usr/bin/env python

from __future__ import print_function

# ImageD11_v1.0 Software for beamline ID11
# Copyright (C) 2008  Jon Wright
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  0211-1307  USA





from ImageD11.refinegrains import refinegrains
import numpy as np

import sys

try:
    flt = sys.argv[1]
    par = sys.argv[2]
    ubi = sys.argv[3]
    tol = float(sys.argv[4])
    if len(sys.argv)>5:
        nbins = int(sys.argv[5])
    else:
        nbins = 30
except:
    print("Usage: %s flt par ubi tol [nbins=30] [omega_slop]"%(sys.argv[0]))
    sys.exit()

if len(sys.argv)>6:
    o=refinegrains(OmFloat=True, OmSlop=float(sys.argv[6]))
else:
    o=refinegrains(OmFloat=False)
                  

o.loadparameters(par)
o.readubis(ubi)
o.loadfiltered(flt)
o.tolerance=tol
o.generate_grains()
o.assignlabels()

import matplotlib.pylab as pl
# indexed peaks only
d = o.scandata[flt]
d.filter(d.labels >= 0)

drlv_bins = np.linspace( 0, tol, nbins )
ng = int(d.labels.max())+1
drlv = np.sqrt( d.drlv2 )
dp5 = [ drlv[d.labels==i] for i in range(ng)]
hl = [ np.histogram(dpi, drlv_bins)[0] for dpi in dp5 ]
print("hl0:",hl[0].shape, drlv_bins.shape)
if drlv_bins.shape[0] != hl[0].shape[0]:
    plotbins = (drlv_bins[1:] + drlv_bins[:-1])/2
    
pl.subplot(211)
for i in range(ng): 
    pl.plot(plotbins,hl[i],label=str(i))

if 1:
    print(" "*10, end=' ')
    for j in range(ng):
        print("%5d"%(j), end=' ')
    print()
    for i in range(len(drlv_bins)-1):
        print("%10.6f"%(drlv_bins[i]), end=' ')
        for j in range(ng):
            print("%5d"%(hl[j][i]), end=' ')
        print()
pl.subplot(212)
pl.hist2d( drlv, d.labels, (drlv_bins, np.arange(-0.5,ng,1.)),vmin=0.5)
pl.colorbar()
pl.ylabel("Grain")
pl.xlabel("drlv")
pl.show()




# Create objects to manipulate - they hold your data
#
from ImageD11 import peakmerge, indexing, transformer
mypeakmerger = peakmerge.peakmerger()
mytransformer = transformer.transformer()
myindexer = indexing.indexer()
#
# Your work starts here:
#
import sys

pars = sys.argv[1]
flt  = sys.argv[2]
gve  = sys.argv[3]
ubi  = sys.argv[4]
mytransformer.loadfileparameters( pars )                               
mytransformer.loadfiltered( flt )
mytransformer.compute_tth_eta( )
mytransformer.addcellpeaks( )
mytransformer.computegv( )
mytransformer.savegv( gve )

myindexer.updateparameters( )
myindexer.parameterobj.set_parameters(  {
    'ds_tol': '0.02', 'minpks': '40', 'uniqueness': '0.5',
    'hkl_tol': '0.05', 'eta_range': '5',
    'ring_1': '2', 'wavelength': '0.26458', 'ring_2': '2', 'cosine_tol': '0.02'} )
myindexer.loadpars( )

myindexer.readgvfile( gve )

myindexer.assigntorings( )

myindexer.find( )
myindexer.scorethem( )
for r1, r2, npks in [
    ( 1,  1, 50 ),
    ( 1,  1, 40 ),
    ( 0,  1, 50 ),
    ( 2,  1, 50 ),
    ( 3,  1, 50 ),
    ( 4,  1, 50 ),
    ( 1 , 1, 50 ),
    ( 0,  1, 40 ),
    ( 2,  1, 40 ),
    ( 3,  1, 40 ),
    ( 4,  1, 40 ),
    ( 1 , 1, 30 ),
    ( 0,  1, 30 ),
    ( 2,  1, 30 ),
    ( 3,  1, 30 ),
    ( 4,  1, 30 ),
    ( 1 , 0, 30 ),
    ( 0,  0, 30 ),
    ( 2,  0, 30 ),
    ( 3,  0, 30 ),
    ( 4,  0, 30 ),
    ]:
    
    myindexer.updateparameters( )
    myindexer.parameterobj.set_parameters(   
    { 'ring_1' : r1,
      'ring_2' : r2,
      'minpks' : npks })
    myindexer.assigntorings( )
    myindexer.loadpars( )
    myindexer.find( )
    myindexer.scorethem( )


myindexer.saveubis( ubi )





from __future__ import print_function
# ImageD11_v0.4 Software for beamline ID11
# Copyright (C) 2015  Jon Wright
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


import numpy as np
from ImageD11 import cImageD11, grain, transform
import unitcell

import math, time, sys, logging

def unit(a): return a/np.sqrt(np.dot(a,a))

def get_tth( ds, wvln):
    return 2 * np.degrees( np.arcsin( ds * wvln / 2 ) )


class indexer:
    """
    A class for searching for orientation matrices

    Base the errors on detector pixels and omega steps
    """
    def __init__(self,
                 transformpars,
                 colfile,
                 ):
        """
        """
        self.transformpars = transformpars
        self.loadcolfile(colfile)        
        self.setcell()
        self.reset()

    def setcell(self):
        cp = [self.transformpars.get("cell_%s"%(s)) for s in
              "_a _b _c alpha beta gamma".split()]
        self.unitcell = unitcell.unitcell(cp, 
                        self.transformpars.get("cell_lattice_[P,A,B,C,I,F,R]"))
        

    def loadcolfile(self, colfile):
        self.cf = colfile
        self.updatecolfile()
        for label in "gx gy gz omega tth eta xl yl zl".split():
            assert label in self.cf.titles, label

    def updatecolfile(self):
       if "xl" not in self.cf.titles:
          if "sc" in self.cf.titles:
             pks = self.cf.sc, self.cf.fc
          elif "xc" in self.cf.titles: 
             pks = self.cf.xc, self.cf.yc
          else:
             raise Exception("peaks file misses xc or sc")
          xl,yl,zl = transform.compute_xyz_lab( pks,
                                                **self.transformpars.parameters)
          self.cf.addcolumn(xl,"xl")
          self.cf.addcolumn(yl,"yl")
          self.cf.addcolumn(zl,"zl")
       peaks_xyz = np.array((self.cf.xl,self.cf.yl,self.cf.zl))
       om = self.cf.omega
       sign = self.transformpars.get("omegasign")
       tth, eta = transform.compute_tth_eta_from_xyz(
          peaks_xyz, om*sign,
          **self.transformpars.parameters)
       self.cf.addcolumn( tth, "tth", )
       self.cf.addcolumn( eta, "eta", )
       gx, gy, gz = transform.compute_g_vectors(
           tth, eta, om*sign,
           wvln  = self.transformpars.get("wavelength"),
           wedge = self.transformpars.get("wedge"),
           chi   = self.transformpars.get("chi") )
       self.cf.addcolumn(gx, "gx")
       self.cf.addcolumn(gy, "gy")
       self.cf.addcolumn(gz, "gz")           
       self.cf.addcolumn( np.sqrt( gx * gx + 
                                   gy * gy + 
                                   gz * gz ),
                          "modg")
       
    
    def reset(self):
        """ when pars change or colfile changes etc """
        if "ring" not in self.cf.titles:
            self.cf.addcolumn( -np.ones(self.cf.nrows), "ring" )
            self.cf.addcolumn( 42*np.ones(self.cf.nrows), "ringerr" )
        if "labels" not in self.cf.titles:
            self.cf.addcolumn( -np.ones(self.cf.nrows), "labels" )

       

    def tthcalc(self, hkls = None):
        if hkls is None:
            dslimit = np.maximum.reduce(self.cf.modg)
            wvln = self.transformpars.get( "wavelength" )
            ttol = self.transformpars.get( "fit_tolerance" ) # tth units
            # at limit this ttol make which ds range?
            tthlim = 2 * np.degrees( np.arcsin( dslimit * wvln / 2 ) )
            dstol  = 2 * np.sin( np.radians( (tthlim + ttol) / 2 ) ) / wvln - dslimit
            self.unitcell.makerings(dslimit+dstol/2, tol = dstol)
        else:
            self.unitcell.peaks = [ (self.unitcell.ds(h),h) for h in hkls ]
            self.unitcell.makerings( dslimit, tol=dstol, newpeaks=False )
        self.unitcell.ringtth = [ 2 * np.degrees( np.arcsin( ds * wvln / 2 ) )
                                  for ds
                                  in self.unitcell.ringds ]

    def assigntorings(self):
        """
        Assign the g-vectors to hkl rings
        """
        dsr = self.unitcell.ringds
        self.cf.ring[:]=-1
        self.cf.ringerr[:]=42.
        tol = self.transformpars.get( "fit_tolerance" )
        for i,tth in enumerate(self.unitcell.ringtth):
           diff = self.cf.tth - tth
           adiff = np.abs(diff)
           mask = (adiff < tol) & (adiff < self.cf.ringerr)
           if mask.sum()>0:
              self.cf.ring[mask] = i
              self.cf.ringerr[mask] = diff[mask]
        # Report on assignments
        print("Ring     (  h,  k,  l) Mult  total indexed to_index  ")
        # try reverse order instead
        dsr = self.unitcell.ringds
        for j in range(len(dsr))[::-1]:
            ind = self.cf.ring == j
            n_indexed  = (self.cf.labels[ind] >  -0.5).sum()
            n_to_index = (self.cf.labels[ind] <  -0.5).sum()
            h=self.unitcell.ringhkls[dsr[j]][0]
            print("Ring %-3d (%3d,%3d,%3d)  %3d  %5d  %5d  %5d"%(\
                j,h[0],h[1],h[2],len(self.unitcell.ringhkls[dsr[j]]),
                     ind.sum(),n_indexed,n_to_index))


    def pairs(self, hkl1, hkl2, cos_tol = 0.02, hkl_tol = 0.05):
        """
        We only look for reflection pairs matching a single hkl pairing
        """
        import time
        start = time.time()
        w = self.transformpars.get( "wavelength" )
        tth1 = get_tth(self.unitcell.ds(hkl1), w) 
        tth2 = get_tth(self.unitcell.ds(hkl2), w)
        tthtol = self.transformpars.get( "fit_tolerance" )
        allinds = np.arange( self.cf.nrows )
        ind1 = allinds[ abs(self.cf.tth - tth1) < tthtol ]
        ind2 = allinds[ abs(self.cf.tth - tth2) < tthtol ]
        angle, cosangle = self.unitcell.anglehkls( hkl1, hkl2 )
        g = np.array( (self.cf.gx, self.cf.gy, self.cf.gz), np.float )
        n = g/self.cf.modg
        gvf = g.T.copy()
        n1 = n[:,ind1]
        n2 = n[:,ind2]
        pairs = []
        j = np.arange( n2.shape[1] )
        # from unitcell.orient
        h1c=unit(np.dot( self.unitcell.B, hkl1 ))
        h2c=unit(np.dot( self.unitcell.B, hkl2 ))
        t1c=unit(h1c)
        t3c=unit(np.cross(h1c,h2c))
        t2c=unit(np.cross(h1c,t3c))
        T_c =  np.array( [t1c, t2c, t3c] )
        T_g = np.zeros((3,3))
        for i,this_n1 in enumerate(n1.T):
           cosa = np.dot( this_n1, n2 )
           goodones = j[abs(cosa - cosangle) < cos_tol]
           # t1 is along g1
           # t2 is plane of both: g1x(g1xg2)
           # t3 is perpendicular to both
           if i%100==0:
              print(i,time.time()-start,len(n1.T))
           for k in goodones:
               this_n2 = n2[:,k]
               T_g[0] = this_n1
               T_g[2] = unit( np.cross( this_n1, this_n2 ) )
               T_g[1] = unit( np.cross( this_n1, T_g[2]) )
               U = np.dot( T_g.T, T_c)
               ubi =  np.dot( U, self.unitcell.B) 
#               npks = cImageD11.score(ubi,gvf,0.1)
               pairs.append( (ind1[i], ind2[k], U, ubi ) )
#               print npks, ubi
        self.pairs=pairs
        print(time.time()-start,"for",len(pairs),n1.shape, n2.shape)
        return pairs



 


if __name__=="__main__":
   from ImageD11.columnfile import columnfile
   from ImageD11.parameters import read_par_file
   import sys
   p = read_par_file(sys.argv[1])
   c = columnfile( sys.argv[2] )
   i = indexer( p, c )
   i.tthcalc()
   i.assigntorings()
   hkl1 = [int(h) for h in sys.argv[3].split(",")]
   hkl2 = [int(h) for h in sys.argv[4].split(",")]
   i.pairs(hkl1, hkl2)




import numpy as np

def norm( v ):
    modv = np.sqrt((np.array(v)*v).sum())
    return v / modv

def project( v ):
    """ project vector onto a pole figure """
    vm = norm( v )
    x = vm[0]/(vm[2] + 1)
    y = vm[1]/(vm[2] + 1)    
    return x,y

def ubi2pf( ubi, v = (0,0,1) ):
    """
    Get the RGB for an inverse pole figure
    """
    hkl = abs( norm( np.dot( ubi, v ) ))
    hkl.sort()
    return project( hkl )

REDPOS = project( (0,0,1 ) )
GREENPOS = project( (0,1,1) )
BLUEPOS = project( (1,1,1) )
CX = (REDPOS[0] + GREENPOS[0] + BLUEPOS[0])/3
CY = (REDPOS[1] + GREENPOS[1] + BLUEPOS[1])/3
dr0 = (REDPOS[0]   - CX)**2 + (REDPOS[1]   - CY)**2
dg0 = (GREENPOS[0] - CX)**2 + (GREENPOS[1] - CY)**2
db0 = (BLUEPOS[0]  - CX)**2 + (BLUEPOS[1]  - CY)**2


def pf2rgb( x, y ):   
    dr = np.arctan2(dr0, ((REDPOS[0]   - x)**2 + (REDPOS[1]   - y)**2))
    dg = np.arctan2(dg0, ((GREENPOS[0] - x)**2 + (GREENPOS[1] - y)**2))
    db = np.arctan2(db0, ((BLUEPOS[0]  - x)**2 + (BLUEPOS[1]  - y)**2))
    m = np.where( dr > dg, dr, dg )
    m = np.where( db > m, db, m )
    rgb = np.array( [dr, dg, db] )
    rgb = rgb / m
    return (rgb*255).astype(np.uint8)
            
def ubi2rgb( ubi ):
    x,y = ubi2pf( ubi )
    return pf2rgb( x, y )







from ImageD11 import cImageD11,  transform

import unittest
import numpy as np

TEST_FORTRAN=False
if TEST_FORTRAN:
    from ImageD11 import fImageD11

class test_compute_gv(unittest.TestCase):
    def setUp(self):
        N = 203

        self.XLYLZL = np.array([ np.linspace(0,2000,N),
                                 np.linspace(10,2010,N),
                                 np.ones(N) ])
        self.XLYLZLT = self.XLYLZL.T.copy()
        self.omega = np.linspace(-180,180,N)

    def runTest(self):
        tth, eta = transform.compute_tth_eta_from_xyz( self.XLYLZL,
                                                       self.omega*self.omegasign,
                                                       t_x=self.t_x,
                                                       t_y=self.t_y,
                                                       t_z=self.t_z,
                                                       wedge=self.wedge,
                                                       chi=self.chi)
        gve1 = transform.compute_g_vectors( tth, eta, self.omega*self.omegasign,
                                            self.wvln, wedge=self.wedge,
                                            chi=self.chi)
        t = np.array( (self.t_x, self.t_y, self.t_z))
        gve2 = np.empty( (len(tth), 3), np.float)
        cImageD11.compute_gv(self.XLYLZLT,
                             self.omega,self.omegasign,
                             self.wvln,
                             self.wedge,self.chi,
                             t,gve2)
        
        #        print t, self.wedge,self.chi
#        for i in range( len(tth)):
#            print i, gve1[:,i],gve1[:,i]-gve2[i],gve1[:,i]-gve3[:,i]
#        print "Test C"
        self.assertTrue( np.allclose(gve1, gve2.T ))
        #        print "Test Fortran"
        if TEST_FORTRAN:
            gve3 = np.empty( (3, len(tth)), np.float, order='F')
            fImageD11.compute_gv(self.XLYLZL,
                                 self.omega,self.omegasign,
                                 self.wvln,
                                 self.wedge,self.chi,
                                 t,
                                 gve3)

        
    def test_5_10(self):
        self.wedge = 5.
        self.chi = 10.
        self.omegasign=1.0
        self.wvln=0.125
        self.t_x=20.
        self.t_y=30.
        self.t_z=40.
        self.runTest()
        
    def test_5_0(self):
        self.wedge = 5.
        self.chi = 0.
        self.omegasign=1.0
        self.wvln=0.125
        self.t_x=20.
        self.t_y=30.
        self.t_z=40.
        self.runTest()

    def test_0_5(self):
        self.wedge = 0.
        self.chi = 5.
        self.omegasign=1.0
        self.wvln=0.125
        self.t_x=20.
        self.t_y=30.
        self.t_z=40.
        self.runTest()

        
    def test_5_10m(self):
        self.wedge = 5.
        self.chi = 10.
        self.omegasign=-1.0
        self.wvln=0.125
        self.t_x=200.
        self.t_y=300.
        self.t_z=40.
        self.runTest()
        
    def test_5_0_tt(self):
        self.wedge = 5.
        self.chi = 0.
        self.omegasign=1.0
        self.wvln=0.125
        self.t_x=200.
        self.t_y=300.
        self.t_z=-400.
        self.runTest()
        
    def test_0_0_t0(self):
        self.wedge = 0.
        self.chi = 0.
        self.omegasign=1.0
        self.wvln=0.125
        self.t_x=00.
        self.t_y=00.
        self.t_z=00.
        self.runTest()

    def test_0_0_t10(self):
        self.wedge = 0.
        self.chi = 0.
        self.omegasign=1.0
        self.wvln=0.125
        self.t_x=200.
        self.t_y=300.
        self.t_z=400.
        self.runTest()

        
if __name__ ==  "__main__":
    unittest.main()

